/**************************************************************/
/* BNCode: Numerical Methods for IVPs of ODEs                 */
/* Copyright (c) 2003-     Tomonori Kouya                     */
/* History:                                                   */
/* Version 0.0.0  2003.2/27: Be Independent from bnc.h        */
/**************************************************************/

#include "bnc.h"

/*************************************************/
/* euler.c: Euler Method for IVP of ODE          */
/*************************************************/
void feuler_1step(FVector, float, FVector, float, void(* func)(FVector, float, FVector), FVector);
void feuler_ex(FILE *, float, FVector, float, FVector, float, void(* func)(FVector, float, FVector), float, float);
void feuler_fs(FILE *, float, FVector, float, FVector, long, void(* func)(FVector, float, FVector));
void deuler_1step(DVector, double, DVector, double, void(* func)(DVector, double, DVector), DVector);
void deuler_ex(FILE *, double, DVector, double, DVector, double, void(* func)(DVector, double, DVector), double, double);
void deuler_fs(FILE *, double, DVector, double, DVector, long, void(* func)(DVector, double, DVector));
#ifdef USE_GMP
void mpf_euler_1step(MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), MPFVector);
void mpf_euler_ex(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t);
void mpf_euler_fs(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, long, void(* func)(MPFVector, mpf_t, MPFVector));
#endif

/*************************************************/
/* rkcoef.c: The above coefficients              */
/*************************************************/
#define ERK_OPTIMAL_2_2   0x0012
#define ERK_CLASSICAL_4_4 0x0014
#define ERK_BUTCHER1_7_6  0x0016

#define IRK_GAUSS_1_2 0x1101
#define IRK_GAUSS_2_4 0x1102
#define IRK_GAUSS_3_6 0x1103
// #define IRK_GAUSS_4_8 0x1004

/* double */
void get_derk_coef(unsigned long, DVector, DMatrix, DVector);
void get_dirk_coef(unsigned long, DVector, DMatrix, DVector);

#ifdef USE_GMP
/* coefficients */
void get_mpf_erk_coef(unsigned long, unsigned long, MPFVector, MPFMatrix, MPFVector);
void get_mpf_irk_coef(unsigned long, unsigned long, MPFVector, MPFMatrix, MPFVector);
#endif

/*************************************************/
/* rk.c : Explicit/Implicit Runge-Kutta Methods  */
/*************************************************/
/* Explicit Runge-Kutta Methods */
void derk_1step(DVector, double, DVector, double, void(* func)(DVector, double, DVector), DVector, DMatrix, DVector, DVector, DVector[]);
void derk_fs(FILE *, double, DVector, double, DVector, long, void(* func)(DVector, double, DVector), unsigned long);
void derk_ex(FILE *, double, DVector, double, DVector, double, void(* func)(DVector, double, DVector), unsigned long, double, double);

/* Implicit Runge-Kutta Methods */
void init_dirk_1step(long int, long int);
void clear_dirk_1step(void);
long int dirk_1step(DVector, double, DVector, double, void(* func)(DVector, double, DVector), void(* jfunc)(DMatrix, double, DVector), DVector, DMatrix, DVector, double, double, long int);
void dirk_fs(FILE *, double, DVector, double, DVector, long, void(* func)(DVector, double, DVector), void(* jfunc)(DMatrix, double, DVector), unsigned long, double, double, long int);
void dirk_ex(FILE *, double, DVector, double, DVector, double, void(* func)(DVector, double, DVector), void(* jfunc)(DMatrix, double, DVector), unsigned long, double, double, long int);

#ifdef USE_GMP
/* Explicit Runge-Kutta Methods */
void mpf_erk_1step(MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), MPFVector, MPFMatrix, MPFVector, MPFVector, MPFVector[]);
void mpf_erk_fs(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, long, void(* func)(MPFVector, mpf_t, MPFVector), unsigned long);
void mpf_erk_ex(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), unsigned long, mpf_t, mpf_t);

/* Implicit Runge-Kutta Methods */
void init_mpf_irk_1step(long int, unsigned long, long int);
void clear_mpf_irk_1step(void);
long int mpf_irk_1step(MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), void(* jfunc)(MPFMatrix, mpf_t, MPFVector), MPFVector, MPFMatrix, MPFVector, mpf_t, mpf_t, long int);
void mpf_irk_fs(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, long, void(* func)(MPFVector, mpf_t, MPFVector), void(* jfunc)(MPFMatrix, mpf_t, MPFVector), unsigned long, mpf_t, mpf_t, long int);
void mpf_irk_ex(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), void(* jfunc)(MPFMatrix, mpf_t, MPFVector), unsigned long, mpf_t, mpf_t, long int);
#endif
/*************************************************/
/* ex_nim.c: Extraporation Method for IVP of ODE */
/*************************************************/
#define MAX_FEX_NIM_STAGE 20
#define MAX_DEX_NIM_STAGE 20

void init_fex_nim(long, long);
void clear_fex_nim(void);
int fex_nim_1step(FVector, float, FVector, float, void(* func)(FVector, float, FVector), float, float);
void fex_nim_fs(FILE *, float, FVector, float, FVector, long, void(* func)(FVector, float, FVector), float, float, long);
void fex_nim(FILE *, float, FVector, float, FVector, float, void(* func)(FVector, float, FVector), float, float, long);

/* inner variables for mpf_ex_nim */
DVector dex_nim_y[MAX_DEX_NIM_STAGE][MAX_DEX_NIM_STAGE];
long int dex_nim_stage;
DVector dex_nim_ytmp[4];
double dex_nim_in_h;

void init_dex_nim(long, long);
void clear_dex_nim(void);
int dex_nim_1step(DVector, double, DVector, double, void(* func)(DVector, double, DVector), double, double);
void dex_nim_fs(FILE *, double, DVector, double, DVector, long, void(* func)(DVector, double, DVector), double, double, long);
//void dex_nim(FILE *, double, DVector, double, DVector, double, void(* func)(DVector, double, DVector), double, double, long);

#ifdef USE_GMP

#define MAX_EX_NIM_STAGE 40

MPFVector ex_nim_y[MAX_EX_NIM_STAGE][MAX_EX_NIM_STAGE];
long int ex_nim_stage;
MPFVector ex_nim_ytmp[4];
mpf_t ex_nim_in_h, ex_nim_mpf_tmp;


void init_mpf_ex_nim(long, long, unsigned long);
void clear_mpf_ex_nim(void);
int mpf_ex_nim_1step(MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t);
void mpf_ex_nim_fs(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, long, void (* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t, long);
//void mpf_ex_nim(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t, long);
#endif

/*************************************************/
/* ex_harmonic.c: Extraporation Method for IVP of ODE */
/*************************************************/
#define MAX_DEX_HARMONIC_STAGE 20

DVector dex_harmonic_y[MAX_DEX_HARMONIC_STAGE][MAX_DEX_HARMONIC_STAGE];
long int dex_harmonic_stage;
DVector dex_harmonic_ytmp[4];
double dex_harmonic_in_h;

double ex_harmonic_den(long int, long int);
void init_dex_harmonic(long int stage, long int dimension);
void clear_dex_harmonic(void);

#ifdef USE_GMP

#define MAX_EX_HARMONIC_STAGE 40

/* inner variables for mpf_ex_harmonic */
MPFVector ex_harmonic_y[MAX_EX_HARMONIC_STAGE][MAX_EX_HARMONIC_STAGE];
long int ex_harmonic_stage;
MPFVector ex_harmonic_ytmp[4];
mpf_t ex_harmonic_in_h, ex_harmonic_mpf_tmp;

void mpf_ex_harmonic_den(mpf_t, long int, long int);
void init_mpf_ex_harmonic(long int, long int, unsigned long);
void clear_mpf_ex_harmonic(void);
int mpf_ex_harmonic_1step(MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t);
void mpf_ex_harmonic_fs(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, long, void (* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t, long);
//void mpf_ex_harmonic(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t, long);
void mpf_ex_harmonic(FILE *, mpf_t, MPFVector, mpf_t, MPFVector, mpf_t, void(* func)(MPFVector, mpf_t, MPFVector), mpf_t, mpf_t, long, void(* ansfunc)(MPFVector, mpf_t), long int);
#endif

/*************************************************/
/* taylor.c: Taylor Expansion Method for IVP of ODE */
/*************************************************/
static long int default_dtaylor_max_degree = 20;

typedef struct {
	long int max_degree; // Maxmum degree of Taylor Expansion 
	long int current_degree; // Current Maximum degree 
	double *coef; // coef[0], coef[1], ..., coef[max_degree]
} double_taylor;

typedef double_taylor *DTaylor;

DTaylor init_dtaylor(void);
DTaylor init_dtaylor_d(double val);
DTaylor init2_dtaylor(long int max_degree);
void free_dtaylor(DTaylor op);
void print_dtaylor(DTaylor op);
void copy_dtaylor(DTaylor rop, DTaylor op);
void set_default_max_degree(long int degree);
void set_const_dtaylor(DTaylor rop, double val);
double dfac(long int n);
void set_exp_dtaylor(DTaylor rop);
void set_sin_dtaylor(DTaylor rop);
void set_cos_dtaylor(DTaylor rop);
void set_current_degree(DTaylor rop);
double eval_dtaylor(DTaylor op, double x);
void neg_dtaylor(DTaylor rop, DTaylor op);
void shift_dtaylor(DTaylor rop, DTaylor op, double x0);
void set_cos2_dtaylor(DTaylor rop, double x0);
void set_sin2_dtaylor(DTaylor rop, double x0);
void add_dtaylor(DTaylor rop, DTaylor op1, DTaylor op2);
void sub_dtaylor(DTaylor rop, DTaylor op1, DTaylor op2);
void smul_dtaylor(DTaylor rop, double sc, DTaylor op);
void sdiv_dtaylor(DTaylor rop, DTaylor op, double sc);
void mul_dtaylor(DTaylor rop, DTaylor op1, DTaylor op2);
void integral_dtaylor(DTaylor rop, DTaylor op, double x0);
void integral2_dtaylor(DTaylor rop, DTaylor op, double x0);
void diff_dtaylor(DTaylor rop, DTaylor op);

#ifdef USE_GMP

/* Taylor Expansion : GMP mpf_t */
static long int default_mpftaylor_max_degree = 20;

typedef struct {
	unsigned long int prec; 
	long int max_degree; // Maxmum degree of Taylor Expansion 
	long int current_degree; // Current Maximum degree 
	mpf_t *coef; // coef[0], coef[1], ..., coef[max_degree]
} mpf_taylor;

typedef mpf_taylor *MPFTaylor;

MPFTaylor init_mpftaylor(void);
MPFTaylor init_mpftaylor_d(mpf_t val);
MPFTaylor init_mpftaylor_str(const char *val);
MPFTaylor init2_mpftaylor(long int max_degree);
void free_mpftaylor(MPFTaylor op);
void print_mpftaylor(MPFTaylor op);
void copy_mpftaylor(MPFTaylor rop, MPFTaylor op);
void set_default_max_degree_mpftaylor(long int degree);
void set_const_mpftaylor(MPFTaylor rop, mpf_t val);
void set_const_mpftaylor_str(MPFTaylor rop, const char *val);
void mpf_fac(mpf_t ret, long int n);
void set_exp_mpftaylor(MPFTaylor rop);
void set_sin_mpftaylor(MPFTaylor rop);
void set_cos_mpftaylor(MPFTaylor rop);
void set_current_degree_mpftaylor(MPFTaylor rop);
void eval_mpftaylor(mpf_t ret, MPFTaylor op, mpf_t x);
void neg_mpftaylor(MPFTaylor rop, MPFTaylor op);
void shift_mpftaylor(MPFTaylor rop, MPFTaylor op, mpf_t x0);
void set_cos2_mpftaylor(MPFTaylor rop, mpf_t x0);
void set_sin2_mpftaylor(MPFTaylor rop, mpf_t x0);
void add_mpftaylor(MPFTaylor rop, MPFTaylor op1, MPFTaylor op2);
void sub_mpftaylor(MPFTaylor rop, MPFTaylor op1, MPFTaylor op2);
void smul_mpftaylor(MPFTaylor rop, mpf_t sc, MPFTaylor op);
void sdiv_mpftaylor(MPFTaylor rop, MPFTaylor op, mpf_t sc);
void mul_mpftaylor(MPFTaylor rop, MPFTaylor op1, MPFTaylor op2);
void integral_mpftaylor(MPFTaylor rop, MPFTaylor op, mpf_t x0);

#endif
