#include <stdio.h>
#include "bnc.h"

/*                       */
/* Explicit Euler Method */
/*                       */

void feuler_1step(FVector y, float x0, FVector y0, float h, void(* func)(FVector y, float x0, FVector y0), FVector ytmp)
{
	/* y := y0 + h * f(x0, y0) */
	func(ytmp, x0, y0);
	cmul2_fvector(ytmp, h);
	add_fvector(y, y0, ytmp);

	return;
}

/*                                           */
/* Explicit Euler Method with fixed stepsize */
/*                                           */
void feuler_fs(FILE *fp, float x, FVector y, float x0, FVector y0, long int div_num, void(* func)(FVector y, float x0, FVector y0))
{
	long int steps, i;
	FVector old_y, tmp_y, lf_tmp;
	float tmp[3], new_x, old_x, in_h;

	/* init */
	old_y = init_fvector(y->dim);
	tmp_y = init_fvector(y->dim);
	lf_tmp = init_fvector(y->dim);

	/* set */
	old_x = x0;
	subst_fvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_h = x - x0;
	if(div_num <= 0)
	{
		fprintf(stderr, "feuler_fs: Number of division is illegal.\n");
		return;
	}
	in_h /= div_num;

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		feuler_1step(tmp_y, old_x, old_y, in_h, func, lf_tmp);

		old_x += in_h;
		subst_fvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%15.7e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%15.7e ", gfvi(old_y, i));
			printf("\n");
		}

	}

	/* finish! */
	subst_fvector(y, old_y);

	printf("OK! feuler_ex has been finished.\n");
	printf("Stepsize  : ");
		 printf("%15.7e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%15.7e", x0);
		 printf(", ");
		 printf("%15.7e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_fvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_fvector(old_y);
	free_fvector(tmp_y);
	free_fvector(lf_tmp);
}

/*                                                                 */
/* Explicit Euler Method with stepsize control using extrapolation */
/*                                                                 */
void feuler_ex(FILE *fp, float x, FVector y, float x0, FVector y0, float max_h, void(* func)(FVector y, float x0, FVector y0), float rtol, float atol)
{
	long int steps, i;
	FVector new_y[2], old_y, tmp_y, lf_tmp;
	float in_max_h, tmp[3], new_x, old_x, in_h, min_h;

	/* init */
	new_y[0] = init_fvector(y->dim);
	new_y[1] = init_fvector(y->dim);
	old_y = init_fvector(y->dim);
	tmp_y = init_fvector(y->dim);
	lf_tmp = init_fvector(y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_fvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = x - x0;
	in_max_h /= 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		in_h = in_max_h;

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		feuler_1step(tmp_y, old_x, old_y, in_h, func, lf_tmp);
		while(1)
		{

			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* new_y[0] := y0 + (h/2) * euler(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * euler(x0 + h/2, new_y[0]) */
			feuler_1step(new_y[0], old_x, old_y, tmp[2], func, lf_tmp);
			tmp[0] = old_x + tmp[2];
			feuler_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, lf_tmp);

			/* calc error */
			sub_fvector(lf_tmp, new_y[1], tmp_y);
			add2_fvector(new_y[1], lf_tmp);

			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			tmp[0] = normi_fvector(lf_tmp);
			tmp[1] = normi_fvector(new_y[1]) * rtol + atol;
			if(tmp[0] <= tmp[1])
				break;

			/* check in_h */
			if(tmp[2] <= 0)
			{
				fprintf(stderr, "feuler_ex: inner stepsize is too small.\n");
				break;
			}
			else
				in_h = tmp[2];
		}

		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_fvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%15.7e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%15.7e ", gfvi(old_y, i));
			printf("\n");
		}

		if(old_x >= x)
			break;
	}

	/* finish! */
	subst_fvector(y, old_y);

	printf("OK! feuler_ex has been finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 printf("%15.7e", min_h);
		 printf(" / ");
		 printf("%15.7e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%15.7e", x0);
		 printf(", ");
		 printf("%15.7e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_fvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_fvector(new_y[0]);
	free_fvector(new_y[1]);
	free_fvector(old_y);
	free_fvector(tmp_y);
	free_fvector(lf_tmp);
}

void deuler_1step(DVector y, double x0, DVector y0, double h, void(* func)(DVector y, double x0, DVector y0), DVector ytmp)
{
	/* y := y0 + h * f(x0, y0) */
	func(ytmp, x0, y0);
	cmul2_dvector(ytmp, h);
	add_dvector(y, y0, ytmp);

	return;
}

/*                                           */
/* Explicit Euler Method with fixed stepsize */
/*                                           */
void deuler_fs(FILE *fp, double x, DVector y, double x0, DVector y0, long int div_num, void(* func)(DVector y, double x0, DVector y0))
{
	long int steps, i;
	DVector old_y, tmp_y, lf_tmp;
	double tmp[3], new_x, old_x, in_h;

	/* init */
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	/* set */
	old_x = x0;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_h = x - x0;
	if(div_num <= 0)
	{
		fprintf(stderr, "deuler_fs: Number of division is illegal.\n");
		return;
	}
	in_h /= div_num;

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		deuler_1step(tmp_y, old_x, old_y, in_h, func, lf_tmp);

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%15.7e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%15.7e ", gdvi(old_y, i));
			printf("\n");
		}

		if(old_x >= x)
			break;
	}

	/* finish! */
	subst_dvector(y, old_y);

	printf("OK! deuler_fs has been finished.\n");
	printf("Stepsize  : ");
		 printf("%25.17e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");

	/* clear */
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);
}
/*                                                                 */
/* Explicit Euler Method with stepsize control using extrapolation */
/*                                                                 */
void deuler_ex(FILE *fp, double x, DVector y, double x0, DVector y0, double max_h, void(* func)(DVector y, double x0, DVector y0), double rtol, double atol)
{
	long int steps, i;
	DVector new_y[2], old_y, tmp_y, lf_tmp;
	double in_max_h, tmp[3], new_x, old_x, in_h, min_h;

	/* init */
	new_y[0] = init_dvector(y->dim);
	new_y[1] = init_dvector(y->dim);
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = x - x0;
	in_max_h /= 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		in_h = in_max_h;

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		deuler_1step(tmp_y, old_x, old_y, in_h, func, lf_tmp);
		while(1)
		{

			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* new_y[0] := y0 + (h/2) * euler(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * euler(x0 + h/2, new_y[0]) */
			deuler_1step(new_y[0], old_x, old_y, tmp[2], func, lf_tmp);
			tmp[0] = old_x + tmp[2];
			deuler_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, lf_tmp);

			/* calc error */
			sub_dvector(lf_tmp, new_y[1], tmp_y);
			add2_dvector(new_y[1], lf_tmp);

			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			tmp[0] = normi_dvector(lf_tmp);
			tmp[1] = normi_dvector(new_y[1]) * rtol + atol;
			if(tmp[0] <= tmp[1])
					break;

			/* check in_h */
			if(tmp[2] <= 0)
			{
				fprintf(stderr, "deuler_ex: inner stepsize is too small.\n");
				break;
			}
			else
				in_h = tmp[2];
		}

		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", gdvi(old_y, i));
			printf("%15.7e", in_h);
			printf("\n");
		}

	}

	/* finish! */
	subst_dvector(y, old_y);

	printf("OK! deuler_fs has been finished.\n");
	printf("Minimum / Maximun stepsize  : ");
		 printf("%25.17e", min_h);
		 printf(" / ");
		 printf("%25.17e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");

	/* clear */
	free_dvector(new_y[0]);
	free_dvector(new_y[1]);
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);
}

#ifdef USE_GMP
void mpf_euler_1step(MPFVector y, mpf_t x0, MPFVector y0, mpf_t h, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), MPFVector ytmp)
{
	/* y := y0 + h * f(x0, y0) */
	func(ytmp, x0, y0);
	cmul2_mpfvector(ytmp, h);
	add_mpfvector(y, y0, ytmp);

	return;
}

/*                                                                 */
/* Explicit Euler Method with stepsize control using extrapolation */
/*                                                                 */
void mpf_euler_ex(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, mpf_t max_h, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), mpf_t rtol, mpf_t atol)
{
	long int steps, i;
	MPFVector new_y[2], old_y, tmp_y, lf_tmp;
	mpf_t in_max_h, tmp[3], new_x, old_x, in_h, min_h;

	/* init */
	new_y[0] = init2_mpfvector(y->dim, y->prec);
	new_y[1] = init2_mpfvector(y->dim, y->prec);
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);

	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));
	mpf_init2(in_max_h, mpf_get_prec(x));
	mpf_init2(min_h, mpf_get_prec(x));

	/* set */
	mpf_set(old_x, x0);
	mpf_set(min_h, max_h);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_max_h, x, x0);
	mpf_div_ui(in_max_h, in_max_h, 2UL);
	if(mpf_cmp(max_h, in_max_h) < 0)
		mpf_set(in_max_h, max_h);

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		mpf_set(in_h, in_max_h);

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		mpf_euler_1step(tmp_y, old_x, old_y, in_h, func, lf_tmp);
		while(1)
		{

			/* h := h / 2 */
			mpf_div_ui(tmp[2], in_h, 2UL);

			/* new_y[0] := y0 + (h/2) * euler(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * euler(x0 + h/2, new_y[0]) */
			mpf_euler_1step(new_y[0], old_x, old_y, tmp[2], func, lf_tmp);
			mpf_add(tmp[0], old_x, tmp[2]);
			mpf_euler_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, lf_tmp);

			/* calc error */
			sub_mpfvector(lf_tmp, new_y[1], tmp_y);
			add2_mpfvector(new_y[1], lf_tmp);

			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			normi_mpfvector(tmp[0], lf_tmp);
			normi_mpfvector(tmp[1], new_y[1]);
			mpf_mul(tmp[1], tmp[1], rtol);
			mpf_add(tmp[1], tmp[1], atol);
			if(mpf_cmp(tmp[0], tmp[1]) <= 0)
					break;

			/* check in_h */
			if(mpf_cmp_si(tmp[2], 0) <= 0)
			{
				fprintf(stderr, "mpf_euler_ex: inner stepsize is too small.\n");
				break;
			}
			else
				mpf_set(in_h, tmp[2]);
		}

		if(mpf_cmp(min_h, in_h) > 0)
			mpf_set(min_h, in_h);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", mpf2double(old_x));
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", mpf2double(gmpfvi(old_y, i)));
			printf("%15.7e ", mpf2double(in_h));
			printf("\n");
		}

		if(mpf_cmp(old_x, x) >= 0)
			break;

		subst_mpfvector(tmp_y, new_y[0]);
	}

	/* finish! */
	subst_mpfvector(y, old_y);

	printf("OK! mpf_euler_ex has been finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 mpf_out_str(stdout, 10, 0, min_h);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 0, max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");

	/* clear */
	free_mpfvector(new_y[0]);
	free_mpfvector(new_y[1]);
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	mpf_clear(in_max_h);
	mpf_clear(min_h);

}

/*                                           */
/* Explicit Euler Method with fixed stepsize */
/*                                           */
void mpf_euler_fs(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, long int div_num, void(* func)(MPFVector y, mpf_t x0, MPFVector y0))
{
	long int steps, i;
	MPFVector old_y, tmp_y, lf_tmp;
	mpf_t tmp[3], new_x, old_x, in_h;

	/* init */
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);

	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));

	/* set */
	mpf_set(old_x, x0);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_h := (x - x0) / div_num) */
	mpf_sub(in_h, x, x0);
	if(div_num <= 0)
	{
		fprintf(stderr, "mpf_euler_fs: Number of division is illegal.\n");
		return;
	}
	mpf_div_ui(in_h, in_h, (unsigned long)div_num);

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		mpf_euler_1step(tmp_y, old_x, old_y, in_h, func, lf_tmp);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", mpf2double(old_x));
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", mpf2double(gmpfvi(old_y, i)));			printf("\n");
		}

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	printf("OK! mpf_euler_fs has been finished.\n");
	printf("Stepsize  : ");
		 mpf_out_str(stdout, 10, 0, in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");

	/* clear */
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);

}
#endif
