/************************************************/
/* ex_nim.c: Extrapolation NIM Method           */
/************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "bncode.h"

/* 2^n - 1 */
unsigned long ex_nim_den(long int n)
{
	unsigned long ans = 1UL;

	while(n-- > 0)
		ans *= 2UL;

	ans--;

	return ans;
}

/* inner variables for mpf_ex_nim */
static FVector fex_nim_y[MAX_FEX_NIM_STAGE][MAX_FEX_NIM_STAGE];
static long int fex_nim_stage;
static FVector fex_nim_ytmp[3];
static float fex_nim_in_h;

void init_fex_nim(long int stage, long int dimension)
{
	long int i, j;

	for(i = 0; i < 3; i++)
		fex_nim_ytmp[i] = init_fvector(dimension);

	fex_nim_stage = stage < (MAX_FEX_NIM_STAGE - 1) ? stage : (MAX_FEX_NIM_STAGE - 1);

	for(i = 0; i < fex_nim_stage; i++)
		for(j = 0; j < fex_nim_stage; j++)
			fex_nim_y[i][j] = init_fvector(dimension);

}

void clear_fex_nim(void)
{
	long int i, j;

	for(i = 0; i < 3; i++)
		free_fvector(fex_nim_ytmp[i]);

	for(i = 0; i < fex_nim_stage; i++)
		for(j = 0; j < fex_nim_stage; j++)
			free_fvector(fex_nim_y[i][j]);

}

/*                          */
/* NIM Extrapolation Method */
/*                          */
int fex_nim_1step(FVector y, float x0, FVector y0, float h, void (* func)(FVector, float, FVector), float rtol, float atol)
{
	static long int i, j, k, istep, jstep, flag = 0, dim;
	static long istep_l;
	static float ftmp, ftmp1;

	istep_l = 2 + 1;
	fex_nim_in_h = h / (float)istep_l;

	for(istep = 0; istep < fex_nim_stage; istep++)
	{

/* First, Using Euler Method */
		func(fex_nim_ytmp[0], x0, y0);
		cmul2_fvector(fex_nim_ytmp[0], fex_nim_in_h);
		add_fvector(fex_nim_y[0][istep], y0, fex_nim_ytmp[0]);

		subst_fvector(fex_nim_ytmp[0], y0);
		subst_fvector(fex_nim_ytmp[1], fex_nim_y[0][istep]);

/* Next, Using Midpoint Rule */
		for(i = 1; i <= istep_l; i++)
		{
			ftmp = x0 + fex_nim_in_h * i;
			func(fex_nim_ytmp[2], ftmp, fex_nim_ytmp[1]);

			cmul2_fvector(fex_nim_ytmp[2], fex_nim_in_h);
			for(j = 0; j < y->dim; j++)
				sfvi(fex_nim_ytmp[2], j, gfvi(fex_nim_ytmp[2], j) * 2);
			add_fvector(fex_nim_y[istep][0], fex_nim_ytmp[0], fex_nim_ytmp[2]);
			if(i == istep_l) /* for smoothing */
				subst_fvector(fex_nim_ytmp[2], fex_nim_ytmp[0]);

			subst_fvector(fex_nim_ytmp[0], fex_nim_ytmp[1]);
			subst_fvector(fex_nim_ytmp[1], fex_nim_y[istep][0]);
		}

/* Smoothing */
		add2_fvector(fex_nim_y[istep][0], fex_nim_ytmp[2]);
		cmul2_fvector(fex_nim_ytmp[0], (float)2.0);
		add2_fvector(fex_nim_y[istep][0], fex_nim_ytmp[0]);
		cmul2_fvector(fex_nim_y[istep][0], (float)0.25);

/* Extrapolation */
		for(jstep = 1; jstep <= istep; jstep++)
		{
			flag = 0;
			for(i = 0; i < y->dim; i++)
			{
				sfvi(fex_nim_ytmp[2], i, gfvi(fex_nim_y[istep][jstep-1], i) - gfvi(fex_nim_y[istep-1][jstep-1], i));
				sfvi(fex_nim_ytmp[2], i, gfvi(fex_nim_ytmp[2], i) / ex_nim_den(jstep*2));
				sfvi(fex_nim_y[istep][jstep], i, gfvi(fex_nim_y[istep][jstep-1], i) + gfvi(fex_nim_ytmp[2], i));

//				if(gfvi(fex_nim_y[istep][jstep], i) != gfvi(fex_nim_y[istep][jstep-1], i))
//					flag = 1;
			}

			sub_fvector(fex_nim_ytmp[1], fex_nim_y[istep][jstep], fex_nim_y[istep][jstep-1]);
			ftmp  = normi_fvector(fex_nim_ytmp[1]);
			ftmp1 = normi_fvector(fex_nim_y[istep][jstep]) * rtol + atol;
//			ftmp  = normi_fvector(fex_nim_y[istep][jstep-1]);
//			ftmp1 = normi_fvector(fex_nim_y[istep][jstep]);
			
			if(ftmp <= ftmp1)
//			if(ftmp == ftmp1)
				flag = 0;
			else
				flag = 1;
			if(flag == 0)
				goto Return;
		}
		fex_nim_in_h /= 2;
		istep_l *= 2;
	}

Return:
	if(istep == fex_nim_stage)
		istep--;
	if(jstep == fex_nim_stage)
		jstep--;
	subst_fvector(y, fex_nim_y[istep][jstep]);

	if(flag == 0) /* equal */
		return 0;
	else		 /* not equal */
		return 1;
}

/*                          */
/* NIM Extrapolation Method */
/*                          */
void fex_nim_fs(FILE *fp, float x, FVector y, float x0, FVector y0, long int div_num, void (* func)(FVector, float, FVector), float rtol, float atol, long int stage)
{
	long int steps, i;
	FVector old_y, tmp_y, lf_tmp;
	float tmp[3], new_x, old_x, in_h;
	
	/* init */
	old_y = init_fvector(y->dim);
	tmp_y = init_fvector(y->dim);
	lf_tmp = init_fvector(y->dim);

	if(stage + 1 > MAX_FEX_NIM_STAGE)
	{
		fprintf(stderr, "fex_nim_fs: Number of stages exceed.\n");
		return;
	}
	init_fex_nim(stage + 1, y->dim);

	/* set */
	old_x = x0;
	subst_fvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_h = x - x0;
	if(div_num <= 0)
	{
		fprintf(stderr, "fex_nim_fs: Number of division is illegal.\n");
		return;
	}

	/* output */
	if(fp != NULL)
	{
		printf("      x        ");
		for(i = 0; i < y->dim; i++)
			printf("  y[%5d]   ", i);
		printf(" stepsize");
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc */
		/* tmp_y := y0 + h * euler(x0, y0) */
		fex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol);

		/* set new x, y */
		old_x += in_h;
		subst_fvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
			fprintf(fp, "%15.7e", old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
				fprintf(fp, "%15.7e", gfvi(old_y, i));
				fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	}

	/* finish! */
	subst_fvector(y, old_y);

	printf("OK! fex_nim has been just finished.\n");
	printf("Stepsize  : ");
		 printf("%15.7e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[");
		 printf("%15.7e", x0);
		 printf(", ");
		 printf("%15.7e", old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_fvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_fvector(old_y);
	free_fvector(tmp_y);
	free_fvector(lf_tmp);

	clear_fex_nim();

}

/*                          */
/* NIM Extrapolation Method */
/*                          */
void fex_nim(FILE *fp, float x, FVector y, float x0, FVector y0, float max_h, void (* func)(FVector, float, FVector), float rtol, float atol, long int stage)
{
	long int steps, i;
	FVector new_y[2], old_y, tmp_y, lf_tmp;
	float tmp[3], new_x, old_x, in_max_h, in_h, min_h;
	
	/* init */
	new_y[0] = init_fvector(y->dim);
	new_y[1] = init_fvector(y->dim);
	old_y = init_fvector(y->dim);
	tmp_y = init_fvector(y->dim);
	lf_tmp = init_fvector(y->dim);

	if(stage + 1 > MAX_FEX_NIM_STAGE)
	{
		fprintf(stderr, "fex_nim: Number of stages exceed.\n");
		return;
	}
	init_fex_nim(stage + 1, y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_fvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = x - x0;
	in_max_h /= 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* output */
	if(fp != NULL)
	{
		printf("      x        ");
		for(i = 0; i < y->dim; i++)
			printf("  y[%5d]   ", i);
		printf(" stepsize");
		printf("\n");
	}

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		in_h = max_h;

		/* check */
		tmp[0] = old_x + in_h;
		if(tmp[0] > x)
			in_h = x - old_x;

		/* calc euler */
		while(1)
		{
			/* tmp_y := y0 + h * euler(x0, y0) */
			if(fex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol) == 0)
				break;

			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* check in_h */
			tmp[0] = old_x + tmp[2];
			if(tmp[0] <= old_x)
			{
				while(tmp[0] <= old_x)
				{
					in_h *= 2;
					tmp[0] = old_x + in_h;
				}
				fex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol);
//				fprintf(stderr, "fex_nim: inner stepsize is too small.\n");
				break;
			}
			else
				in_h = tmp[2];
		}

		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_fvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
			fprintf(fp, "%15.7e", old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
				fprintf(fp, "%15.7e", gfvi(old_y, i));
				fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

		if(old_x >= x)
			break;

	}

	/* finish! */
	subst_fvector(y, old_y);

	printf("OK! fex_nim has been just finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 printf("%15.7e", min_h);
		 printf(" / ");
		 printf("%15.7e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[");
		 printf("%15.7e", x0);
		 printf(", ");
		 printf("%15.7e", old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_fvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_fvector(new_y[0]);
	free_fvector(new_y[1]);
	free_fvector(old_y);
	free_fvector(tmp_y);
	free_fvector(lf_tmp);

	clear_fex_nim();

}

/* double */

void init_dex_nim(long int stage, long int dimension)
{
	long int i, j;

	for(i = 0; i < 4; i++)
		dex_nim_ytmp[i] = init_dvector(dimension);

	dex_nim_stage = stage < (MAX_DEX_NIM_STAGE - 1) ? stage : (MAX_DEX_NIM_STAGE - 1);

	for(i = 0; i < dex_nim_stage; i++)
		for(j = 0; j < dex_nim_stage; j++)
			dex_nim_y[i][j] = init_dvector(dimension);

}

void clear_dex_nim(void)
{
	long int i, j;

	for(i = 0; i < 4; i++)
		free_dvector(dex_nim_ytmp[i]);

	for(i = 0; i < dex_nim_stage; i++)
		for(j = 0; j < dex_nim_stage; j++)
			free_dvector(dex_nim_y[i][j]);

}

/*                          */
/* NIM Extrapolation Method */
/*                          */
int dex_nim_1step(DVector y, double x0, DVector y0, double h, void (* func)(DVector, double, DVector), double rtol, double atol)
{
	static long int i, j, k, istep, jstep, flag = 0, dim;
	static long istep_l;
	static double ftmp, ftmp1;

	istep_l = 2;
	dex_nim_in_h = h / (double)(istep_l);

	for(istep = 0; istep < dex_nim_stage; istep++)
	{

/* First, Using Euler Method */
		func(dex_nim_ytmp[0], x0, y0);
		cmul2_dvector(dex_nim_ytmp[0], dex_nim_in_h);
		add_dvector(dex_nim_y[0][istep], y0, dex_nim_ytmp[0]);

		subst_dvector(dex_nim_ytmp[0], y0);
		subst_dvector(dex_nim_ytmp[1], dex_nim_y[0][istep]);

/* Next, Using Midpoint Rule */
		for(i = 1; i <= istep_l; i++)
		{
			ftmp = x0 + dex_nim_in_h * i;
			func(dex_nim_ytmp[2], ftmp, dex_nim_ytmp[1]);

			cmul2_dvector(dex_nim_ytmp[2], dex_nim_in_h);
			for(j = 0; j < y->dim; j++)
				sdvi(dex_nim_ytmp[2], j, gdvi(dex_nim_ytmp[2], j) * 2);
			add_dvector(dex_nim_y[istep][0], dex_nim_ytmp[0], dex_nim_ytmp[2]);
			if(i == istep_l) /* for smoothing */
				subst_dvector(dex_nim_ytmp[2], dex_nim_ytmp[0]);
			subst_dvector(dex_nim_ytmp[0], dex_nim_ytmp[1]);
			subst_dvector(dex_nim_ytmp[1], dex_nim_y[istep][0]);
		}

/* Smoothing */
		add2_dvector(dex_nim_y[istep][0], dex_nim_ytmp[2]);
		cmul2_dvector(dex_nim_ytmp[0], 2.0);
		add2_dvector(dex_nim_y[istep][0], dex_nim_ytmp[0]);
		cmul2_dvector(dex_nim_y[istep][0], 0.25);

/* Extrapolation */
		for(jstep = 1; jstep <= istep; jstep++)
		{
			flag = 0;
			for(i = 0; i < y->dim; i++)
			{
				sdvi(dex_nim_ytmp[2], i, gdvi(dex_nim_y[istep][jstep-1], i) - gdvi(dex_nim_y[istep-1][jstep-1], i));
				sdvi(dex_nim_ytmp[2], i, gdvi(dex_nim_ytmp[2], i) / ex_nim_den(jstep*2));
				sdvi(dex_nim_y[istep][jstep], i, gdvi(dex_nim_y[istep][jstep-1], i) + gdvi(dex_nim_ytmp[2], i));

//				if(gdvi(dex_nim_y[istep][jstep], i) != gdvi(dex_nim_y[istep][jstep-1], i))
//					flag = 1;
			}

			sub_dvector(dex_nim_ytmp[1], dex_nim_y[istep][jstep], dex_nim_y[istep][jstep-1]);
			ftmp  = normi_dvector(dex_nim_ytmp[1]);
			ftmp1 = normi_dvector(dex_nim_y[istep][jstep]) * rtol + atol;
			if(ftmp <= ftmp1)
				flag = 0;
			else
				flag = 1;
			if(flag == 0)
				goto Return;
		}
		istep_l *= 2;
		dex_nim_in_h = h / (double)(istep_l);
	}

Return:
	if(istep == dex_nim_stage)
		istep--;
	if(jstep == dex_nim_stage)
		jstep--;
	subst_dvector(y, dex_nim_y[istep][jstep]);

	if(flag == 0) /* equal */
		return 0;
//		return istep;
	else		 /* not equal */
		return 1;
}

/*                          */
/* NIM Extrapolation Method */
/*                          */
void dex_nim_fs(FILE *fp, double x, DVector y, double x0, DVector y0, long int div_num, void (* func)(DVector, double, DVector), double rtol, double atol, long int stage)
{
	long int steps, i, num_stage;
	DVector old_y, tmp_y, lf_tmp;
	double tmp[3], new_x, old_x, in_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();

	/* init */
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	if(stage + 1 > MAX_DEX_NIM_STAGE)
	{
		fprintf(stderr, "dex_nim_fs: Number of stages exceed.\n");
		return;
	}
	init_dex_nim(stage + 1, y->dim);

	/* set */
	old_x = x0;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_h := (x - x0) / div_num */
	if(div_num <= 0)
	{
		fprintf(stderr, "dex_nim_fs: Number of division is illegal.\n");
		return;
	}
	in_h = (x - x0) / div_num;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* tmp_y := y0 + h * ex_nim(x0, y0) */
		num_stage = dex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol);

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
			fprintf(fp, "%25.17e", old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
				fprintf(fp, "%25.17e", gdvi(old_y, i));
				fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	}

	/* finish! */
	subst_dvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! dex_nim_fs has been just finished.\n");
	printf("Number of Stages: %d\n", stage);
	printf("Relative Tol/Absolute Tol: ");
		 printf("%.5e", rtol);
		 printf(" / ");
		 printf("%.5e", atol);
		 printf("\n");
	printf("Stepsize  : ");
		 printf("%25.17e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);

	clear_dex_nim();

}

/*                          */
/* NIM Extrapolation Method */
/*                          */
void dex_nim(FILE *fp, double x, DVector y, double x0, DVector y0, double max_h, void (* func)(DVector, double, DVector), double rtol, double atol, long int stage, void (* ansfunc)(DVector, double))
{
	long int steps, i, conv_flag;
	DVector new_y[2], old_y, tmp_y, lf_tmp;
	double tmp[3], new_x, old_x, in_max_h, in_h, min_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();
	
	/* init */
	new_y[0] = init_dvector(y->dim);
	new_y[1] = init_dvector(y->dim);
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	if(stage + 1 > MAX_DEX_NIM_STAGE)
	{
		fprintf(stderr, "dex_nim: Number of stages exceed.\n");
		return;
	}
	init_dex_nim(stage + 1, y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = (x - x0) / 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("\n");
	}

	/* main loop */
	in_h = max_h;
//	in_h = 1.0e-10;
	for(steps = 0; ; steps++)
	{
		/* set default interval */
//		in_h = max_h;
		in_h *= 2;
		if(in_h > max_h)
			in_h = max_h;

		/* check */
		tmp[0] = old_x + in_h;
		if(tmp[0] > x)
			in_h = x - old_x;

		/* calc euler */
		while(1)
		{
			/* tmp_y := y0 + h * ex(x0, y0) */
			if(dex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol) == 0)
				break;

			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* check in_h */
			tmp[0] = old_x + tmp[2];
			if(tmp[0] <= old_x)
			{
				while(tmp[0] <= old_x)
				{
					in_h *= 2;
					tmp[0] = old_x + in_h;
				}
				dex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol);
				fprintf(stderr, "dex_harmonic: inner stepsize is too small.(%15.7e)\n", in_h);
				break;
			}
			else
				in_h = tmp[2];
		}
		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
			ansfunc(tmp_y, old_x);
			fprintf(fp, "%25.17e", old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
//				fprintf(fp, "%25.17e", gdvi(old_y, i));
				fprintf(fp, "%10.7e", fabs(gdvi(old_y, i) - gdvi(tmp_y, i))/fabs(gdvi(tmp_y, i)));
				fprintf(fp, " ");
			}
			fprintf(fp, "%15.7e ", in_h);
			fprintf(fp, "\n");
		}

		if(old_x >= x)
			break;

	}

	/* finish! */
	subst_dvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! dex_nim has been just finished.\n");
	printf("Number of Stages: %d\n", stage);
	printf("Relative Tol/Absolute Tol: ");
		 printf("%.5e", rtol);
		 printf(" / ");
		 printf("%.5e", atol);
		 printf("\n");
	printf("Minimum/Maxmum stepsize  : ");
		 printf("%15.7e", min_h);
		 printf(" / ");
		 printf("%15.7e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 printf("%15.7e", x0);
		 printf(", ");
		 printf("%15.7e", old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_dvector(new_y[0]);
	free_dvector(new_y[1]);
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);

	clear_dex_nim();

}

#ifdef USE_GMP
/* inner variables for mpf_ex_nim */
void init_mpf_ex_nim(long int stage, long int dimension, unsigned long prec)
{
	long int i, j;

	mpf_init2(ex_nim_in_h, prec); mpf_init2(ex_nim_mpf_tmp, prec);
	for(i = 0; i < 4; i++)
		ex_nim_ytmp[i] = init2_mpfvector(dimension, prec);

	ex_nim_stage = stage < (MAX_EX_NIM_STAGE - 1) ? stage : (MAX_EX_NIM_STAGE - 1);

	for(i = 0; i < ex_nim_stage; i++)
		for(j = 0; j < ex_nim_stage; j++)
			ex_nim_y[i][j] = init2_mpfvector(dimension, prec);

}

void clear_mpf_ex_nim(void)
{
	long int i, j;

	mpf_clear(ex_nim_in_h); mpf_clear(ex_nim_mpf_tmp);
	for(i = 0; i < 4; i++)
		free_mpfvector(ex_nim_ytmp[i]);

	for(i = 0; i < ex_nim_stage; i++)
		for(j = 0; j < ex_nim_stage; j++)
			free_mpfvector(ex_nim_y[i][j]);

}

/*                          */
/* NIM Extrapolation Method */
/*                          */
int mpf_ex_nim_1step(MPFVector y, mpf_t x0, MPFVector y0, mpf_t h, void (* func)(MPFVector, mpf_t, MPFVector), mpf_t rtol, mpf_t atol)
{
	long int i, j, k, istep, jstep, flag = 0, dim;
	long istep_l;

	istep_l = 2;
	mpf_div_ui(ex_nim_in_h, h, (unsigned long)istep_l);

	for(istep = 0; istep < ex_nim_stage; istep++)
	{

/* First, Using Euler Method */
		func(ex_nim_ytmp[0], x0, y0);
		cmul2_mpfvector(ex_nim_ytmp[0], ex_nim_in_h);
		add_mpfvector(ex_nim_y[0][istep], y0, ex_nim_ytmp[0]);

		subst_mpfvector(ex_nim_ytmp[0], y0);
		subst_mpfvector(ex_nim_ytmp[1], ex_nim_y[0][istep]);

/* Next, Using Midpoint Rule */
		for(i = 1; i <= istep_l; i++)
		{
			mpf_mul_ui(ex_nim_mpf_tmp, ex_nim_in_h, (unsigned long)i);
			mpf_add(ex_nim_mpf_tmp, x0, ex_nim_mpf_tmp);
			func(ex_nim_ytmp[2], ex_nim_mpf_tmp, ex_nim_ytmp[1]);

			cmul2_mpfvector(ex_nim_ytmp[2], ex_nim_in_h);
			for(j = 0; j < y->dim; j++)
				mpf_mul_ui(
					get_mpfvector_i(ex_nim_ytmp[2], j),
					get_mpfvector_i(ex_nim_ytmp[2], j),
					(unsigned long)2
				);
			add_mpfvector(ex_nim_y[istep][0], ex_nim_ytmp[0], ex_nim_ytmp[2]);
			if(i == istep_l)
				subst_mpfvector(ex_nim_ytmp[2], ex_nim_ytmp[0]);
			subst_mpfvector(ex_nim_ytmp[0], ex_nim_ytmp[1]);
			subst_mpfvector(ex_nim_ytmp[1], ex_nim_y[istep][0]);
		}
/* Smoothing */
		add2_mpfvector(ex_nim_y[istep][0], ex_nim_ytmp[2]);
		mpf_set_ui(ex_nim_mpf_tmp, 2UL);
		cmul2_mpfvector(ex_nim_ytmp[0], ex_nim_mpf_tmp);
		add2_mpfvector(ex_nim_y[istep][0], ex_nim_ytmp[0]);
		mpf_set_ui(ex_nim_mpf_tmp, 1UL);
		mpf_div_ui(ex_nim_mpf_tmp, ex_nim_mpf_tmp, 4UL);
		cmul2_mpfvector(ex_nim_y[istep][0], ex_nim_mpf_tmp);

/* Extrapolation */
		for(jstep = 1; jstep <= istep; jstep++)
		{
			flag = 0;
			for(i = 0; i < y->dim; i++)
			{
				mpf_sub(
					get_mpfvector_i(ex_nim_ytmp[2], i),
					get_mpfvector_i(ex_nim_y[istep][jstep-1], i),
					get_mpfvector_i(ex_nim_y[istep-1][jstep-1], i)
				);
				mpf_div_ui(
					get_mpfvector_i(ex_nim_ytmp[2], i),
				 	get_mpfvector_i(ex_nim_ytmp[2], i),
				 	ex_nim_den(jstep*2)
				);
				mpf_add(
					get_mpfvector_i(ex_nim_y[istep][jstep], i),
					get_mpfvector_i(ex_nim_y[istep][jstep-1], i),
					get_mpfvector_i(ex_nim_ytmp[2], i)
				);
			
//				if(mpf_cmp(get_mpfvector_i(ex_nim_y[istep][jstep], i), get_mpfvector_i(ex_nim_y[istep][jstep-1], i) ) != 0)
//					flag = 1;
			}

			/* (ex_nim_ytmp[0], 0) := || ex_nim_y[istep][jstep] - ex_nim_y[istep][jstep-1] || * rtol + atol */
			sub_mpfvector(ex_nim_ytmp[1], ex_nim_y[istep][jstep], ex_nim_y[istep][jstep-1]);
			normi_mpfvector(get_mpfvector_i(ex_nim_ytmp[0], 0), ex_nim_ytmp[1]);
	
			/* ex_nim_tmp := || ex_nim_y[istep][jstep] || */
			normi_mpfvector(ex_nim_mpf_tmp, ex_nim_y[istep][jstep]);
			mpf_mul(ex_nim_mpf_tmp, ex_nim_mpf_tmp, rtol);
			mpf_add(ex_nim_mpf_tmp, ex_nim_mpf_tmp, atol);

			/* || ex_nim_y[istep][jstep] - ex_nim_y[istep][jstep-1]  || <= || ex_nim_y[istep][jstep] || * rtol + atol */
			if(mpf_cmp(gmpfvi(ex_nim_ytmp[0], 0), ex_nim_mpf_tmp) <= 0)
				flag = 0;
			else
				flag = 1;
		//	print_mpfvector(ex_y[istep][jstep]);
			if(flag == 0)
			{
//				fflush(stdout); printf("return! %ld\n", istep);
				goto Return;
			}
		}
		mpf_div_ui(ex_nim_in_h, ex_nim_in_h, (unsigned long)2);
		istep_l *= 2;
//		printf("%d(%d) ", istep_l, ex_nim_stage);
	}

Return:
	if(istep == ex_nim_stage)
		istep--;
	if(jstep == ex_nim_stage)
		jstep--;
	subst_mpfvector(y, ex_nim_y[istep][jstep]);

	if(flag == 0) /* equal */
		return 0;
	else		 /* not equal */
		return 1;
}

/*                          */
/* NIM Extrapolation Method */
/* Stepsize fixed           */
void mpf_ex_nim_fs(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, long int div_num, void (* func)(MPFVector, mpf_t, MPFVector), mpf_t rtol, mpf_t atol, long int stage)
{
	long int steps, i;
	MPFVector old_y, tmp_y, lf_tmp;
	mpf_t tmp[3], new_x, old_x, in_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();
	
	/* init */
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));

	if(stage + 1 > MAX_EX_NIM_STAGE)
	{
		fprintf(stderr, "mpf_ex_nim_fs: Number of stages exceed.\n");
		return;
	}
	init_mpf_ex_nim(stage + 1, y->dim, y->prec);

	/* set */
	mpf_set(old_x, x0);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_h := (x - x0) / div_num */
	mpf_sub(in_h, x, x0);
	if(div_num <= 0)
	{
		fprintf(stderr, "mpf_ex_nim_fs: A number of division is illegal.\n");
		return;
	}
	mpf_div_ui(in_h, in_h, div_num);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc euler */
		/* tmp_y := y0 + h * euler(x0, y0) */
		mpf_ex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
		if(fp != NULL)
		{
//			mpf_out_str(fp, 10, 0, old_x);
			fprintf(fp, "%25.17e", mpf2double(old_x));
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
//				mpf_out_str(fp, 10, 0, gmpfvi(old_y, i));
				fprintf(fp, "%25.17e", mpf_get_d(gmpfvi(old_y, i)));
				fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! mpf_ex_nim_fs has been just finished.\n");
	printf("Number of Stages: %d\n", stage);
	printf("Relative Tol/Absolute Tol: ");
		 mpf_out_str(stdout, 10, 5, rtol);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 5, atol);
		 printf("\n");
	printf("Stepsize  : ");
		 mpf_out_str(stdout, 10, 0, in_h);
//		 printf(", Number of division: %ld ", div_num);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);

	clear_mpf_ex_nim();

}

/*                          */
/* NIM Extrapolation Method */
/*                          */
void mpf_ex_nim(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, mpf_t max_h, void (* func)(MPFVector, mpf_t, MPFVector), mpf_t rtol, mpf_t atol, long int stage, void (* ansfunc)(MPFVector, mpf_t), long int itimes_iv)
{
	long int steps, i;
	MPFVector new_y[2], old_y, tmp_y, lf_tmp;
	mpf_t tmp[3], new_x, old_x, in_max_h, in_h, min_h;
	double start_time, end_time;

	/* time */
	start_time = get_secv();
	
	/* init */
	new_y[0] = init2_mpfvector(y->dim, y->prec);
	new_y[1] = init2_mpfvector(y->dim, y->prec);
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));
	mpf_init2(in_max_h, mpf_get_prec(x));
	mpf_init2(min_h, mpf_get_prec(x));

	if(stage + 1 > MAX_EX_NIM_STAGE)
	{
		fprintf(stderr, "mpf_ex_nim: Number of stages exceed.\n");
		return;
	}
	init_mpf_ex_nim(stage + 1, y->dim, y->prec);

	/* set */
	mpf_set(old_x, x0);
	mpf_set(min_h, max_h);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_max_h, x, x0);
	mpf_div_ui(in_max_h, in_max_h, 2UL);
	if(mpf_cmp(max_h, in_max_h) < 0)
		mpf_set(in_max_h, max_h);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("\n");
		printf("DEBUG...\n"); fflush(stdout);
	}
		printf("DEBUG...\n"); fflush(stdout);

	/* main loop */
	mpf_set(in_h, max_h);
	for(steps = 0; ; steps++)
	{
		printf("DEBUG...\n"); fflush(stdout);
		/* set default interval */
		mpf_mul_ui(in_h, in_h, 2UL);
		if(mpf_cmp(in_h, max_h) > 0)
			mpf_set(in_h, max_h);

		/* check */
		mpf_add(tmp[0], old_x, in_h);
		if(mpf_cmp(tmp[0], x) > 0)
			mpf_sub(in_h, x, old_x);

		/* calc euler */
		while(1)
		{
			/* tmp_y := y0 + h * euler(x0, y0) */
			if(mpf_ex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol) == 0)
				break;

			/* h := h / 2 */
			mpf_div_ui(tmp[2], in_h, 2UL);

			/* check in_h */
			mpf_add(tmp[0], old_x, tmp[2]);
			if(mpf_cmp(tmp[0], old_x) <= 0)
			{
				while(mpf_cmp(tmp[0], old_x) <= 0)
				{
					mpf_mul_ui(in_h, in_h, 2UL);
					mpf_add(tmp[0], old_x, in_h);
				}
				mpf_ex_nim_1step(tmp_y, old_x, old_y, in_h, func, rtol, atol);
				fprintf(stderr, "mpf_ex_nim: inner stepsize is too small.\n");
				break;
			}
			else
				mpf_set(in_h, tmp[2]);
		}

		if(mpf_cmp(min_h, in_h) > 0)
			mpf_set(min_h, in_h);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* x, y1, y2, ..., yn */
//		if(fp != NULL)
		if((fp != NULL) && ((steps % itimes_iv) == 0))
		{
//			mpf_out_str(fp, 10, 0, old_x);
			fprintf(fp, "%25.17e", mpf2double(old_x));
			ansfunc(tmp_y, old_x);
			fprintf(fp, " ");
			for(i = 0; i < old_y->dim; i++)
			{
//				mpf_out_str(fp, 10, 0, gmpfvi(old_y, i));
//				fprintf(fp, "%25.17e", mpf_get_d(gmpfvi(old_y, i) - gmpfvi(tmp_y, i)));
				mpf_sub(tmp[0], gmpfvi(old_y, i), gmpfvi(tmp_y, i));
				mpf_div(tmp[0], tmp[0], gmpfvi(tmp_y, i));
				mpf_abs(tmp[0], tmp[0]);
				mpf_out_str(fp, 10, 7, tmp[0]);
				fprintf(fp, " ");
			}
			fprintf(fp, "%15.7e ", mpf2double(in_h));
			fprintf(fp, "\n");
		}

		if(mpf_cmp(old_x, x) >= 0)
			break;

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	/* time */
	end_time = get_secv();

	printf("OK! mpf_ex_nim has been just finished.\n");
	printf("Number of Stages: %d\n", stage);
	printf("Relative Tol/Absolute Tol: ");
		 mpf_out_str(stdout, 10, 5, rtol);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 5, atol);
		 printf("\n");
	printf("Minimum/Maxmum stepsize  : ");
		 mpf_out_str(stdout, 10, 0, min_h);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 0, max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Computational Time(sec): %f\n", end_time - start_time);
	printf("Integral interval : ");
		 printf("[");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf("]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_mpfvector(new_y[0]);
	free_mpfvector(new_y[1]);
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	mpf_clear(in_max_h);
	mpf_clear(min_h);

	clear_mpf_ex_nim();

}
#endif
