/************************************************/
/* rk.c : Explicit & Implicit Runge-Kutta Methods*/
/************************************************/
#include <stdio.h>
#include "bncode.h"

/* Explicit Runge-Kutta Method */
void derk_1step(DVector y, double x0, DVector y0, double h, void(* func)(DVector y, double x0, DVector y0), DVector coef_c, DMatrix coef_a, DVector coef_w, DVector lf_tmp, DVector k[])
{
	static long int i, j;
	static double tmp;

/* y := y0 + h * (w1 * k1 + w2 * k2 + ... wm * km) */

/* k1 */
	func(k[0], x0, y0);

/* k2...km */
	for(i = 1; i < coef_w->dim; i++)
	{
		cmul_dvector(y, gdmij(coef_a, i-1, 0), k[0]);
		for(j = 1; j < i; j++)
		{
			cmul_dvector(lf_tmp, gdmij(coef_a, i-1, j), k[j]);
			add2_dvector(y, lf_tmp);
		}
		cmul2_dvector(y, h);
		add2_dvector(y, y0); /* y0 + h * (a_i1 * k1 + ... + ai4 * k4) */
		tmp = x0 + gdvi(coef_c, i-1) * h;
		func(k[i], tmp, y);
	}

/* y_n+1 := y_n + h * \sum^4_{j=1} w_j * k_j */
	cmul_dvector(y, gdvi(coef_w, 0), k[0]);
	for(i = 1; i < coef_w->dim; i++)
	{
		cmul_dvector(lf_tmp, gdvi(coef_w, i), k[i]);
		add2_dvector(y, lf_tmp);
	}
	cmul2_dvector(y, h);
	add2_dvector(y, y0);

	return ;
}

/*                                                 */
/* Explicit Runge-Kutta Method with fixed stepsize */
/*                                                 */
void derk_fs(FILE *fp, double x, DVector y, double x0, DVector y0, long int div_num, void(* func)(DVector y, double x0, DVector y0), unsigned long coef_type)
{
	long int steps, i;
	DVector old_y, tmp_y, lf_tmp, k[7];
	double in_h, tmp[3], new_x, old_x;
	DVector coef_c, coef_w;
	DMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case ERK_OPTIMAL_2_2:
			coef_c = init_dvector(1);
			coef_w = init_dvector(2);
			coef_a = init_dmatrix(1, 1);
			break;
		
		case ERK_CLASSICAL_4_4:
			coef_c = init_dvector(3);
			coef_w = init_dvector(4);
			coef_a = init_dmatrix(3, 3);
			break;

		case ERK_BUTCHER1_7_6:
			coef_c = init_dvector(6);
			coef_w = init_dvector(7);
			coef_a = init_dmatrix(6, 6);
			break;

		default:
			fprintf(stderr, "derk_fs: Such a coefficient set not exist.\n");
			return;
	}

	get_derk_coef(coef_type, coef_c, coef_a, coef_w);

//	printf("c: \n"); print_dvector(coef_c);
//	printf("w: \n"); print_dvector(coef_w);
//	printf("a: \n"); print_dmatrix(coef_a);

	/* init */
	for(i = 0; i < coef_w->dim; i++)
		k[i] = init_dvector(y->dim);
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	/* set */
	old_x = x0;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_h = x - x0;
	if(div_num <= 0)
	{
		fprintf(stderr, "derk_fs: Number of division is illegal.\n");
		return;
	}
	in_h /= div_num;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc */

		/* tmp_y := y0 + h * erk(x0, y0) */
		derk_1step(tmp_y, old_x, old_y, in_h, func, coef_c, coef_a, coef_w, lf_tmp, k);

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", gdvi(old_y, i));
			printf("\n");
		}

	}

	/* finish! */
	subst_dvector(y, old_y);

	printf("OK! derk_fs has been finished.\n");
	printf("Stepsize  : ");
		 printf("%25.17e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");

	/* clear */
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);
	for(i = 0; i < coef_w->dim; i++)
		free_dvector(k[i]);
	free_dvector(coef_c);
	free_dvector(coef_w);
	free_dmatrix(coef_a);

}

/*                                                                       */
/* Explicit Runge-Kutta Method with stepsize control using extrapolation */
/*                                                                       */
void derk_ex(FILE *fp, double x, DVector y, double x0, DVector y0, double max_h, void(* func)(DVector y, double x0, DVector y0), unsigned long coef_type, double rtol, double atol)
{
	long int steps, i;
	DVector new_y[2], old_y, tmp_y, lf_tmp, k[7];
	double in_max_h, tmp[3], new_x, old_x, in_h, min_h;
	DVector coef_c, coef_w;
	DMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case ERK_OPTIMAL_2_2:
			coef_c = init_dvector(1);
			coef_w = init_dvector(2);
			coef_a = init_dmatrix(1, 1);
			break;
		
		case ERK_CLASSICAL_4_4:
			coef_c = init_dvector(3);
			coef_w = init_dvector(4);
			coef_a = init_dmatrix(3, 3);
			break;

		case ERK_BUTCHER1_7_6:
			coef_c = init_dvector(6);
			coef_w = init_dvector(7);
			coef_a = init_dmatrix(6, 6);
			break;

		default:
			fprintf(stderr, "clear_derk_coef: Such a coefficient set not exist.\n");
			return;
	}

	get_derk_coef(coef_type, coef_c, coef_a, coef_w);

//	printf("c: \n"); print_dvector(coef_c);
//	printf("w: \n"); print_dvector(coef_w);
//	printf("a: \n"); print_dmatrix(coef_a);

	/* init */
	for(i = 0; i < coef_w->dim; i++)
		k[i] = init_dvector(y->dim);
	new_y[0] = init_dvector(y->dim);
	new_y[1] = init_dvector(y->dim);
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = x - x0;
	in_max_h /= 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("\n");
	}

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		in_h = in_max_h;

		/* calc */

		/* tmp_y := y0 + h * erk(x0, y0) */
		derk_1step(tmp_y, old_x, old_y, in_h, func, coef_c, coef_a, coef_w, lf_tmp, k);
		while(1)
		{
			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* new_y[0] := y0 + (h/2) * euler(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * euler(x0 + h/2, new_y[0]) */
			derk_1step(new_y[0], old_x, old_y, tmp[2], func, coef_c, coef_a, coef_w, lf_tmp, k);

			tmp[0] = old_x + tmp[2];
			derk_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, coef_c, coef_a, coef_w, lf_tmp, k);

			/* calc error */
			sub_dvector(lf_tmp, new_y[1], tmp_y);
			add2_dvector(new_y[1], lf_tmp);
	
			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			tmp[0] = normi_dvector(lf_tmp);
			tmp[1] = normi_dvector(new_y[1]) * rtol + atol;
			if(tmp[0] <= tmp[1])
					break;

			/* check in_h */
			if(tmp[2] <= 0)
			{
				fprintf(stderr, "derk_ex: inner stepsize is too small.\n");
				break;
			}

			in_h = tmp[2];
			subst_dvector(tmp_y, new_y[0]);

		}

		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", gdvi(old_y, i));
			printf("%15.7e ", in_h);
			printf("\n");
		}

		if(old_x >= x)
			break;

	}

	/* finish! */
	subst_dvector(y, old_y);

	printf("OK! derk_ex has been finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 printf("%25.17e", min_h);
		 printf(" / ");
		 printf("%25.17e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");

	/* clear */
	free_dvector(new_y[0]);
	free_dvector(new_y[1]);
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);
	for(i = 0; i < coef_w->dim; i++)
		free_dvector(k[i]);
	free_dvector(coef_c);
	free_dvector(coef_w);
	free_dmatrix(coef_a);

}

#ifdef USE_GMP

/* Explicit Runge-Kutta Method */
void mpf_erk_1step(MPFVector y, mpf_t x0, MPFVector y0, mpf_t h, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), MPFVector coef_c, MPFMatrix coef_a, MPFVector coef_w, MPFVector lf_tmp, MPFVector k[])
{
	long int i, j;

/* y := y0 + h * (w1 * k1 + w2 * k2 + ... wm * km) */

/* k1 */
	func(k[0], x0, y0);

/* k2...km */
	for(i = 1; i < coef_w->dim; i++)
	{
		cmul_mpfvector(y, gmpfmij(coef_a, i-1, 0), k[0]);
		for(j = 1; j < i; j++)
		{
			cmul_mpfvector(lf_tmp, gmpfmij(coef_a, i-1, j), k[j]);
			add2_mpfvector(y, lf_tmp);
		}
		cmul2_mpfvector(y, h);
		add2_mpfvector(y, y0); /* y0 + h * (a_i1 * k1 + ... + ai4 * k4) */
		mpf_mul(get_mpfvector_i(lf_tmp, 0), gmpfvi(coef_c, i-1), h);
		mpf_add(get_mpfvector_i(lf_tmp, 0), x0, gmpfvi(lf_tmp, 0));
		func(k[i], gmpfvi(lf_tmp, 0), y);
	}

/* y_n+1 := y_n + h * \sum^4_{j=1} w_j * k_j */
	cmul_mpfvector(y, gmpfvi(coef_w, 0), k[0]);
	for(i = 1; i < coef_w->dim; i++)
	{
		cmul_mpfvector(lf_tmp, gmpfvi(coef_w, i), k[i]);
		add2_mpfvector(y, lf_tmp);
	}
	cmul2_mpfvector(y, h);
	add2_mpfvector(y, y0);

	return ;
}

/*                                                 */
/* Explicit Runge-Kutta Method with fixed stepsize */
/*                                                 */
void mpf_erk_fs(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, long int div_num, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), unsigned long coef_type)
{
	long int steps, i;
	MPFVector old_y, tmp_y, lf_tmp, k[7];
	mpf_t tmp[3], new_x, old_x, in_h;
	MPFVector coef_c, coef_w;
	MPFMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case ERK_OPTIMAL_2_2:
			coef_c = init2_mpfvector(1, y->prec);
			coef_w = init2_mpfvector(2, y->prec);
			coef_a = init2_mpfmatrix(1, 1, y->prec);
			break;
		
		case ERK_CLASSICAL_4_4:
			coef_c = init2_mpfvector(3, y->prec);
			coef_w = init2_mpfvector(4, y->prec);
			coef_a = init2_mpfmatrix(3, 3, y->prec);
			break;

		case ERK_BUTCHER1_7_6:
			coef_c = init2_mpfvector(6, y->prec);
			coef_w = init2_mpfvector(7, y->prec);
			coef_a = init2_mpfmatrix(6, 6, y->prec);
			break;

		default:
			fprintf(stderr, "mpf_erk_fs: Such a coefficient set not exist.\n");
			return;
	}

	get_mpf_erk_coef(coef_type, y->prec, coef_c, coef_a, coef_w);

//	printf("c: \n"); print_mpfvector(coef_c);
//	printf("w: \n"); print_mpfvector(coef_w);
//	printf("a: \n"); print_mpfmatrix(coef_a);

	/* init */
	for(i = 0; i < coef_w->dim; i++)
		k[i] = init2_mpfvector(y->dim, y->prec);
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));

	/* set */
	mpf_set(old_x, x0);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_h, x, x0);
	if(div_num <= 0)
	{
		fprintf(stderr, "mpf_erk_fs: Number of division is illegal.\n");
		return;
	}
	mpf_div_ui(in_h, in_h, (unsigned long)div_num);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc */

		/* tmp_y := y0 + h * erk(x0, y0) */
		mpf_erk_1step(tmp_y, old_x, old_y, in_h, func, coef_c, coef_a, coef_w, lf_tmp, k);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", mpf2double(old_x));
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", mpf2double(gmpfvi(old_y, i)));
			printf("\n");
		}

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	printf("OK! mpf_erk_fs has been finished.\n");
	printf("Stepsize  : ");
		 mpf_out_str(stdout, 10, 0, in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	for(i = 0; i < coef_w->dim; i++)
		free_mpfvector(k[i]);
	free_mpfvector(coef_c);
	free_mpfvector(coef_w);
	free_mpfmatrix(coef_a);

}

/*                                                                       */
/* Explicit Runge-Kutta Method with stepsize control using extrapolation */
/*                                                                       */
void mpf_erk_ex(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, mpf_t max_h, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), unsigned long coef_type, mpf_t rtol, mpf_t atol)
{
	long int steps, i;
	MPFVector new_y[2], old_y, tmp_y, lf_tmp, k[7];
	mpf_t in_max_h, tmp[3], new_x, old_x, in_h, min_h;
	MPFVector coef_c, coef_w;
	MPFMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case ERK_OPTIMAL_2_2:
			coef_c = init2_mpfvector(1, y->prec);
			coef_w = init2_mpfvector(2, y->prec);
			coef_a = init2_mpfmatrix(1, 1, y->prec);
			break;
		
		case ERK_CLASSICAL_4_4:
			coef_c = init2_mpfvector(3, y->prec);
			coef_w = init2_mpfvector(4, y->prec);
			coef_a = init2_mpfmatrix(3, 3, y->prec);
			break;

		case ERK_BUTCHER1_7_6:
			coef_c = init2_mpfvector(6, y->prec);
			coef_w = init2_mpfvector(7, y->prec);
			coef_a = init2_mpfmatrix(6, 6, y->prec);
			break;

		default:
			fprintf(stderr, "clear_mpf_erk_coef: Such a coefficient set not exist.\n");
			return;
	}

	get_mpf_erk_coef(coef_type, y->prec, coef_c, coef_a, coef_w);

//	printf("c: \n"); print_mpfvector(coef_c);
//	printf("w: \n"); print_mpfvector(coef_w);
//	printf("a: \n"); print_mpfmatrix(coef_a);

	/* init */
	for(i = 0; i < coef_w->dim; i++)
		k[i] = init2_mpfvector(y->dim, y->prec);
	new_y[0] = init2_mpfvector(y->dim, y->prec);
	new_y[1] = init2_mpfvector(y->dim, y->prec);
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));
	mpf_init2(in_max_h, mpf_get_prec(x));
	mpf_init2(min_h, mpf_get_prec(x));

	/* set */
	mpf_set(old_x, x0);
	mpf_set(min_h, max_h);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_max_h, x, x0);
	mpf_div_ui(in_max_h, in_max_h, 2UL);
	if(mpf_cmp(max_h, in_max_h) < 0)
		mpf_set(in_max_h, max_h);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("\n");
	}

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		mpf_set(in_h, in_max_h);

		/* calc */

		/* tmp_y := y0 + h * erk(x0, y0) */
		mpf_erk_1step(tmp_y, old_x, old_y, in_h, func, coef_c, coef_a, coef_w, lf_tmp, k);
		while(1)
		{
			/* h := h / 2 */
			mpf_div_ui(tmp[2], in_h, 2UL);

			/* new_y[0] := y0 + (h/2) * euler(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * euler(x0 + h/2, new_y[0]) */
			mpf_erk_1step(new_y[0], old_x, old_y, tmp[2], func, coef_c, coef_a, coef_w, lf_tmp, k);

			mpf_add(tmp[0], old_x, tmp[2]);
			mpf_erk_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, coef_c, coef_a, coef_w, lf_tmp, k);

			/* calc error */
			sub_mpfvector(lf_tmp, new_y[1], tmp_y);
			add2_mpfvector(new_y[1], lf_tmp);

			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			normi_mpfvector(tmp[0], lf_tmp);
			normi_mpfvector(tmp[1], new_y[1]);
			mpf_mul(tmp[1], tmp[1], rtol);
			mpf_add(tmp[1], tmp[1], atol);
			if(mpf_cmp(tmp[0], tmp[1]) <= 0)
					break;

			/* check in_h */
			if(mpf_cmp_si(tmp[2], 0) <= 0)
			{
				fprintf(stderr, "mpf_erk: inner stepsize is too small.\n");
				break;
			}

			mpf_set(in_h, tmp[2]);
			subst_mpfvector(tmp_y, new_y[0]);

//			printf("%10.7e %25.17e %25.17e\n", mpf2double(in_h), mpf2double(gmpfvi(lf_tmp, 0)), mpf2double(gmpfvi(new_y[1], 0)));
		}

		if(mpf_cmp(min_h, in_h) > 0)
			mpf_set(min_h, in_h);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", mpf2double(old_x));
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", mpf2double(gmpfvi(old_y, i)));
			printf("%15.7e ", mpf2double(in_h));
			printf("\n");
		}

		if(mpf_cmp(old_x, x) >= 0)
			break;

//		printf("step, in_h: %ld, %25.17e, %25.17e\n", steps, mpf2double(old_x), mpf2double(gmpfvi(old_y,0))); fflush(stdout);
//		printf("step, in_h: %ld\n", steps);
//		fflush(stdout);
	}

	/* finish! */
	subst_mpfvector(y, old_y);

	printf("OK! mpf_erk has been finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 mpf_out_str(stdout, 10, 0, min_h);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 0, max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");
//	mpf_set(x, old_x);

	/* clear */
	free_mpfvector(new_y[0]);
	free_mpfvector(new_y[1]);
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	mpf_clear(in_max_h);
	mpf_clear(min_h);
	for(i = 0; i < coef_w->dim; i++)
		free_mpfvector(k[i]);
	free_mpfvector(coef_c);
	free_mpfvector(coef_w);
	free_mpfmatrix(coef_a);

}
#endif

/* for mpf_irk */
static DMatrix dirk_jacobim;
static DMatrix dirk_jacobi;
static DVector dirk_k_old;
static DVector dirk_k_tmp;
static DVector dirk_k_new;
static DVector dirk_ytmp;
static DVector dirk_lf_tmp;
static DVector dirk_lf_tmp1;

/* init mpf_irk_1step */
void init_dirk_1step(long int dimension, long int steps)
{
	long int jdim;

	if(steps <= 0)
	{
		fprintf(stderr, "Error: Steps of Implicit Runge-Kutta methhod is invalid.\n");
		return;
	}

	jdim = dimension * steps;

	dirk_jacobim = init_dmatrix(jdim, jdim);
	dirk_jacobi = init_dmatrix(dimension, dimension);
	dirk_k_new = init_dvector(jdim);
	dirk_k_tmp = init_dvector(jdim);
	dirk_k_old = init_dvector(jdim);
	dirk_ytmp = init_dvector(dimension);
	dirk_lf_tmp = init_dvector(dimension);
	dirk_lf_tmp1 = init_dvector(dimension);
}

/* init mpf_irk */
void clear_dirk_1step(void)
{
	free_dmatrix(dirk_jacobim);
	free_dmatrix(dirk_jacobi);
	free_dvector(dirk_k_old);
	free_dvector(dirk_k_tmp);
	free_dvector(dirk_k_new);
	free_dvector(dirk_ytmp);
	free_dvector(dirk_lf_tmp);
	free_dvector(dirk_lf_tmp1);
}


void x_set_dirk_jacobim(DMatrix jm, DVector coef_c, DMatrix coef_a, double x0, DVector y0, void(* jfunc)(DMatrix jacobi, double x0, DVector y0), double h, DVector km, DMatrix jacobi, DVector ytmp)
{
	static long int i, j, ki, kj;
	static double xtmp, tmp;

	for(i = 0; i < coef_a->row_dim; i++)
	{
		/* xtmp := x0 + c[i] * h */
		xtmp = x0 + h * gdvi(coef_c, i);

		/* ytmp := y0 + h * \sum^n_{j=1} a[i,j] * km) */
		set0_dvector(ytmp);
		for(j = 0; j < coef_a->col_dim; j++)
		{
			for(ki = 0; ki < y0->dim; ki++)
				sdvi(ytmp, ki, gdvi(ytmp, ki) + gdmij(coef_a, i, j) * gdvi(km, j * y0->dim + ki));
		}
		cmul2_dvector(ytmp, h);
		add2_dvector(ytmp, y0);

		/* jacobi(x0 + c[i] * h, y0 + \sum^{j=1} a[i,j] * k_j) */
		jfunc(jacobi, xtmp, ytmp);

		for(j = 0; j < coef_a->col_dim; j++)
		{
			for(ki = 0; ki < jacobi->row_dim; ki++)
			{
				for(kj = 0; kj < jacobi->col_dim; kj++)
				{
					tmp = -h * gdmij(coef_a, i, j) * gdmij(jacobi, ki, kj);
					sdmij(jm, i * jacobi->row_dim + ki, j * jacobi->col_dim + kj, tmp);
				}
			}
		}
	}

	for(i = 0; i < jm->row_dim; i++)
		sdmij(jm, i, i, 1 + gdmij(jm, i, i));

}

long int dirk_1step(DVector y, double x0, DVector y0, double h, void(* func)(DVector y, double x0, DVector y0), void(* jfunc)(DMatrix jacobi, double x0, DVector y0), DVector coef_c, DMatrix coef_a, DVector coef_w, double abs_tol, double rel_tol, long int maxtimes)
{
	long int i, j, k, times;
	double tmp, tmp1;

/* y := y0 + h * (w1 * k1 + w2 * k2 + ... wm * km) */

/* k1...km */
	for(i = 0; i < coef_w->dim; i++)
	{
		tmp = x0 + gdvi(coef_c, i) * h;
		func(dirk_lf_tmp, tmp, y0);
		for(j = 0; j < y->dim; j++)
			sdvi(dirk_k_old, i * y->dim + j, gdvi(dirk_lf_tmp, j));
	}

	/* simplified Newton Method */

	x_set_dirk_jacobim(dirk_jacobim, coef_c, coef_a, x0, y0, jfunc, h, dirk_k_old, dirk_jacobi, dirk_ytmp);
	DLUdecomp(dirk_jacobim);

	/* Main loop */
	for(times = 0; times < maxtimes; times++)
	{
		/* ktmpm := k_i - func(x0 + c_i h, y0 + h (a_i1 k_1 + ... + a_im k_m) ) */
		for(i = 0; i < coef_a->row_dim; i++)
		{
			/* x0 + c_i * h */
			tmp = x0 + gdvi(coef_c, i) * h;

			/* y0 + h * (a_i1 * k_1 + ... + a_im * k_m) */
			set0_dvector(dirk_lf_tmp);
			for(j = 0; j < coef_a->col_dim; j++)
			{
				for(k = 0; k < y->dim; k++)
					sdvi(dirk_lf_tmp, k, gdvi(dirk_lf_tmp, k) + gdmij(coef_a, i, j) * gdvi(dirk_k_old, j * y->dim + k));
			}
			cmul2_dvector(dirk_lf_tmp, h);
			add2_dvector(dirk_lf_tmp, y0);

			func(dirk_lf_tmp1, tmp, dirk_lf_tmp);

			for(j = 0; j < y->dim; j++)
				sdvi(dirk_k_tmp, i * y->dim + j, gdvi(dirk_k_old, i * y->dim + j) - gdvi(dirk_lf_tmp1, j));
		}
		/* jacobi(x0)^(-1) * func(x) */
		SolveDLS(dirk_k_new, dirk_jacobim, dirk_k_tmp);

		/* check */
		/* |func(x)/dfunc(x)| <= abs_eps + |x_old| * rel_eps */
		tmp  = abs_tol + rel_tol * normi_dvector(dirk_k_old);
		tmp1 = normi_dvector(dirk_k_new);

		/* set next step */
		sub_dvector(dirk_k_old, dirk_k_old, dirk_k_new);

		if(tmp1 <= tmp)
			break;
	}

	/* return */
	if(times >= maxtimes)
		fprintf(stderr, "dirk_1step: innner iteration is not convergent.\n");

/* y_n+1 := y_n + h * \sum^m_{j=1} w_j * k_j */
	set0_dvector(y);
	for(i = 0; i < coef_w->dim; i++)
	{
		for(j = 0; j < y->dim; j++)
			sdvi(y, j, gdvi(y, j) + gdvi(coef_w, i) * gdvi(dirk_k_old, i * y->dim + j));
	}
	cmul2_dvector(y, h);
	add2_dvector(y, y0);

	return times;
}

/*                                                 */
/* Implicit Runge-Kutta Method with fixed stepsize */
/*                                                 */
void dirk_fs(FILE *fp, double x, DVector y, double x0, DVector y0, long int div_num, void(* func)(DVector y, double x0, DVector y0), void(* jfunc)(DMatrix ymat, double x0, DVector y0), unsigned long coef_type, double rtol, double atol, long int maxtimes)
{
	long int steps, i, itimes;
	DVector old_y, tmp_y, lf_tmp;
	double tmp[3], new_x, old_x, in_h;
	DVector coef_c, coef_w;
	DMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case IRK_GAUSS_1_2:
			coef_c = init_dvector(1);
			coef_w = init_dvector(1);
			coef_a = init_dmatrix(1, 1);
			steps = 1;
			break;
		
		case IRK_GAUSS_2_4:
			coef_c = init_dvector(2);
			coef_w = init_dvector(2);
			coef_a = init_dmatrix(2, 2);
			steps = 2;
			break;

		case IRK_GAUSS_3_6:
			coef_c = init_dvector(3);
			coef_w = init_dvector(3);
			coef_a = init_dmatrix(3, 3);
			steps = 3;
			break;

		default:
			fprintf(stderr, "dirk_fs: Such a coefficient set not exist.\n");
			return;
	}

	get_dirk_coef(coef_type, coef_c, coef_a, coef_w);

	init_dirk_1step(y->dim, steps);

//	printf("c: \n"); print_dvector(coef_c);
//	printf("w: \n"); print_dvector(coef_w);
//	printf("a: \n"); print_dmatrix(coef_a);

	/* init */
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	/* set */
	old_x = x0;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_h := (x - x0) / div_num */
	in_h = x - x0;
	if(div_num <= 0)
	{
		fprintf(stderr, "dirk_fs: Number of division is illegal.\n");
		return;
	}
	in_h /= div_num;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("times\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* calc */

		/* tmp_y := y0 + h * irk(x0, y0) */
		itimes = dirk_1step(tmp_y, old_x, old_y, in_h, func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", gdvi(old_y, i));
			printf(" %ld", itimes);
			printf("\n");
		}
	}

	/* finish! */
	subst_dvector(y, old_y);

	printf("OK! dirk_fs has been finished.\n");
	printf("Stepsize  : ");
		 printf("%25.17e", in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");

	/* clear */
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);
	free_dvector(coef_c);
	free_dvector(coef_w);
	free_dmatrix(coef_a);

}
/*                                                                       */
/* Implicit Runge-Kutta Method with stepsize control using extrapolation */
/*                                                                       */
void dirk_ex(FILE *fp, double x, DVector y, double x0, DVector y0, double max_h, void(* func)(DVector y, double x0, DVector y0), void(* jfunc)(DMatrix ymat, double x0, DVector y0), unsigned long coef_type, double rtol, double atol, long int maxtimes)
{
	long int steps, i, itimes[2];
	DVector new_y[2], old_y, tmp_y, lf_tmp;
	double in_max_h, tmp[3], new_x, old_x, in_h, min_h;
	DVector coef_c, coef_w;
	DMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case IRK_GAUSS_1_2:
			coef_c = init_dvector(1);
			coef_w = init_dvector(1);
			coef_a = init_dmatrix(1, 1);
			steps = 1;
			break;
		
		case IRK_GAUSS_2_4:
			coef_c = init_dvector(2);
			coef_w = init_dvector(2);
			coef_a = init_dmatrix(2, 2);
			steps = 2;
			break;

		case IRK_GAUSS_3_6:
			coef_c = init_dvector(3);
			coef_w = init_dvector(3);
			coef_a = init_dmatrix(3, 3);
			steps = 3;
			break;

		default:
			fprintf(stderr, "dirk_ex: Such a coefficient set not exist.\n");
			return;
	}

	get_dirk_coef(coef_type, coef_c, coef_a, coef_w);

	init_dirk_1step(y->dim, steps);

//	printf("c: \n"); print_dvector(coef_c);
//	printf("w: \n"); print_dvector(coef_w);
//	printf("a: \n"); print_dmatrix(coef_a);

	/* init */
	new_y[0] = init_dvector(y->dim);
	new_y[1] = init_dvector(y->dim);
	old_y = init_dvector(y->dim);
	tmp_y = init_dvector(y->dim);
	lf_tmp = init_dvector(y->dim);

	/* set */
	old_x = x0;
	min_h = max_h;
	subst_dvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	in_max_h = (x - x0) / 2;
	if(max_h < in_max_h)
		in_max_h = max_h;

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("times(h, h/2)\n");
	}

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		in_h = in_max_h;

		/* calc */

		/* tmp_y := y0 + h * irk(x0, y0) */
		dirk_1step(tmp_y, old_x, old_y, in_h, func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);
		while(1)
		{
			/* h := h / 2 */
			tmp[2] = in_h / 2;

			/* new_y[0] := y0 + (h/2) * irk(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * irk(x0 + h/2, new_y[0]) */
			itimes[0] = dirk_1step(new_y[0], old_x, old_y, tmp[2], func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);

			tmp[0] = old_x + tmp[2];
			itimes[1] = dirk_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);

			/* calc error */
			sub_dvector(lf_tmp, new_y[1], tmp_y);
			add2_dvector(new_y[1], lf_tmp);
	
			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			tmp[0] = normi_dvector(lf_tmp);
			tmp[1] = normi_dvector(new_y[1]) * rtol + atol;
			if(tmp[0] <= tmp[1])
					break;

			/* check in_h */
			if(tmp[2] <= 0)
			{
				fprintf(stderr, "dirk_ex: inner stepsize is too small.\n");
				break;
			}

			in_h = tmp[2];
			subst_dvector(tmp_y, new_y[0]);

		}

		if(min_h > in_h)
			min_h = in_h;

		/* set new x, y */
		old_x += in_h;
		subst_dvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", old_x);
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", gdvi(old_y, i));
			printf("%15.7e ", in_h);
			printf(" %ld %ld", itimes[0], itimes[1]);
			printf("\n");
		}

		if(old_x >= x)
			break;

	}

	/* finish! */
	subst_dvector(y, old_y);

	printf("OK! dirk_ex has been finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 printf("%25.17e", min_h);
		 printf(" / ");
		 printf("%25.17e", max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 printf("%25.17e", x0);
		 printf(", ");
		 printf("%25.17e", old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_dvector(y);
		 printf("\n");

	/* clear */
	free_dvector(new_y[0]);
	free_dvector(new_y[1]);
	free_dvector(old_y);
	free_dvector(tmp_y);
	free_dvector(lf_tmp);
	free_dvector(coef_c);
	free_dvector(coef_w);
	free_dmatrix(coef_a);

}

#ifdef USE_GMP
/* for mpf_irk */
static MPFMatrix mpf_irk_jacobim;
static MPFMatrix mpf_irk_jacobi;
static MPFVector mpf_irk_k_old;
static MPFVector mpf_irk_k_tmp;
static MPFVector mpf_irk_k_new;
static MPFVector mpf_irk_ytmp;
static MPFVector mpf_irk_lf_tmp;
static MPFVector mpf_irk_lf_tmp1;

/* init mpf_irk_1step */
void init_mpf_irk_1step(long int dimension, unsigned long prec, long int steps)
{
	long int jdim;

	if(steps <= 0)
	{
		fprintf(stderr, "Error: Steps of Implicit Runge-Kutta methhod is invalid.\n");
		return;
	}

	jdim = dimension * steps;

	mpf_irk_jacobim = init2_mpfmatrix(jdim, jdim, prec);
	mpf_irk_jacobi = init2_mpfmatrix(dimension, dimension, prec);
	mpf_irk_k_new = init2_mpfvector(jdim, prec);
	mpf_irk_k_tmp = init2_mpfvector(jdim, prec);
	mpf_irk_k_old = init2_mpfvector(jdim, prec);
	mpf_irk_ytmp = init2_mpfvector(dimension, prec);
	mpf_irk_lf_tmp = init2_mpfvector(dimension, prec);
	mpf_irk_lf_tmp1 = init2_mpfvector(dimension, prec);
}

/* init mpf_irk */
void clear_mpf_irk_1step(void)
{
	free_mpfmatrix(mpf_irk_jacobim);
	free_mpfmatrix(mpf_irk_jacobi);
	free_mpfvector(mpf_irk_k_old);
	free_mpfvector(mpf_irk_k_tmp);
	free_mpfvector(mpf_irk_k_new);
	free_mpfvector(mpf_irk_ytmp);
	free_mpfvector(mpf_irk_lf_tmp);
	free_mpfvector(mpf_irk_lf_tmp1);
}


void x_set_mpf_irk_jacobim(MPFMatrix jm, MPFVector coef_c, MPFMatrix coef_a, mpf_t x0, MPFVector y0, void(* jfunc)(MPFMatrix jacobi, mpf_t x0, MPFVector y0), mpf_t h, MPFVector km, MPFMatrix jacobi, MPFVector ytmp)
{
	long int i, j, ki, kj;
	mpf_t xtmp, tmp;
//	MPFMatrix jacobi;
//	MPFVector ytmp;

	mpf_init2(tmp, jm->prec);
	mpf_init2(xtmp, jm->prec);
//	jacobi = init2_mpfmatrix(y0->dim, y0->dim, jm->prec);
//	ytmp = init2_mpfvector(y0->dim, jm->prec);

	for(i = 0; i < coef_a->row_dim; i++)
	{
		/* xtmp := x0 + c[i] * h */
		mpf_mul(xtmp, h, gmpfvi(coef_c, i));
		mpf_add(xtmp, x0, xtmp);

		/* ytmp := y0 + h * \sum^n_{j=1} a[i,j] * km) */
		set0_mpfvector(ytmp);
		for(j = 0; j < coef_a->col_dim; j++)
		{
			for(ki = 0; ki < y0->dim; ki++)
			{
				mpf_mul(tmp, gmpfmij(coef_a, i, j), gmpfvi(km, j * y0->dim + ki));
				mpf_add(gmpfvi(ytmp, ki), gmpfvi(ytmp, ki), tmp);
			}
		}
		cmul2_mpfvector(ytmp, h);
		add2_mpfvector(ytmp, y0);

		/* jacobi(x0 + c[i] * h, y0 + \sum^{j=1} a[i,j] * k_j) */
		jfunc(jacobi, xtmp, ytmp);

		for(j = 0; j < coef_a->col_dim; j++)
		{
			for(ki = 0; ki < jacobi->row_dim; ki++)
			{
				for(kj = 0; kj < jacobi->col_dim; kj++)
				{
					mpf_mul(tmp, h, gmpfmij(coef_a, i, j));
					mpf_mul(tmp, tmp, gmpfmij(jacobi, ki, kj));
					mpf_neg(tmp, tmp); /* tmp := -tmp */
					smpfmij(jm, i * jacobi->row_dim + ki, j * jacobi->col_dim + kj, tmp);
				}
			}
		}
	}

	mpf_set_ui(tmp, 1UL);
	for(i = 0; i < jm->row_dim; i++)
		mpf_add(gmpfmij(jm, i, i), tmp, gmpfmij(jm, i, i));

	mpf_clear(tmp);
	mpf_clear(xtmp);
//	free_mpfmatrix(jacobi);
//	free_mpfvector(ytmp);

}

long int mpf_irk_1step(MPFVector y, mpf_t x0, MPFVector y0, mpf_t h, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), void(* jfunc)(MPFMatrix jacobi, mpf_t x0, MPFVector y0), MPFVector coef_c, MPFMatrix coef_a, MPFVector coef_w, mpf_t abs_tol, mpf_t rel_tol, long int maxtimes)
{
	long int i, j, k, times;
	mpf_t tmp, tmp1;

	mpf_init2(tmp, y->prec);
	mpf_init2(tmp1, y->prec);
/* y := y0 + h * (w1 * k1 + w2 * k2 + ... wm * km) */

/* k1...km */
	for(i = 0; i < coef_w->dim; i++)
	{
		mpf_mul(tmp, gmpfvi(coef_c, i), h);
		mpf_add(tmp, x0, tmp);
		func(mpf_irk_lf_tmp, tmp, y0);
		for(j = 0; j < y->dim; j++)
			smpfvi(mpf_irk_k_old, i * y->dim + j, gmpfvi(mpf_irk_lf_tmp, j));
	}

	/* simplified Newton Method */

	x_set_mpf_irk_jacobim(mpf_irk_jacobim, coef_c, coef_a, x0, y0, jfunc, h, mpf_irk_k_old, mpf_irk_jacobi, mpf_irk_ytmp);
	MPFLUdecomp(mpf_irk_jacobim);

	/* Main loop */
	for(times = 0; times < maxtimes; times++)
	{
		/* ktmpm := k_i - func(x0 + c_i h, y0 + h (a_i1 k_1 + ... + a_im k_m) ) */
		for(i = 0; i < coef_a->row_dim; i++)
		{
			/* x0 + c_i * h */
			mpf_mul(tmp, gmpfvi(coef_c, i), h);
			mpf_add(tmp, x0, tmp);

			/* y0 + h * (a_i1 * k_1 + ... + a_im * k_m) */
			set0_mpfvector(mpf_irk_lf_tmp);
			for(j = 0; j < coef_a->col_dim; j++)
			{
				for(k = 0; k < y->dim; k++)
				{
					mpf_mul(tmp1, gmpfmij(coef_a, i, j), gmpfvi(mpf_irk_k_old, j * y->dim + k));
					mpf_add(gmpfvi(mpf_irk_lf_tmp, k), gmpfvi(mpf_irk_lf_tmp, k), tmp1);
				}
			}
			cmul2_mpfvector(mpf_irk_lf_tmp, h);
			add2_mpfvector(mpf_irk_lf_tmp, y0);

			func(mpf_irk_lf_tmp1, tmp, mpf_irk_lf_tmp);

			for(j = 0; j < y->dim; j++)
			{
				mpf_sub(tmp, gmpfvi(mpf_irk_k_old, i * y->dim + j), gmpfvi(mpf_irk_lf_tmp1, j));
				smpfvi(mpf_irk_k_tmp, i * y->dim + j, tmp);
			}
		}
		/* jacobi(x0)^(-1) * func(x) */
		SolveMPFLS(mpf_irk_k_new, mpf_irk_jacobim, mpf_irk_k_tmp);

		/* check */
		/* |func(x)/dfunc(x)| <= abs_eps + |x_old| * rel_eps */
		normi_mpfvector(tmp, mpf_irk_k_old);
		mpf_mul(tmp, tmp, rel_tol);
		mpf_add(tmp, tmp, abs_tol);
		normi_mpfvector(tmp1, mpf_irk_k_new);

		/* set next step */
		sub_mpfvector(mpf_irk_k_old, mpf_irk_k_old, mpf_irk_k_new);

		if(mpf_cmp(tmp1, tmp) <= 0)
			break;
	}

	/* return */
	if(times >= maxtimes)
		fprintf(stderr, "mpf_irk_1step: innner iteration is not convergent.\n");

/* y_n+1 := y_n + h * \sum^m_{j=1} w_j * k_j */
	set0_mpfvector(y);
	for(i = 0; i < coef_w->dim; i++)
	{
		for(j = 0; j < y->dim; j++)
		{
			mpf_mul(tmp1, gmpfvi(coef_w, i), gmpfvi(mpf_irk_k_old, i * y->dim + j));
			mpf_add(gmpfvi(y, j), gmpfvi(y, j), tmp1);
		}
	}
	cmul2_mpfvector(y, h);
	add2_mpfvector(y, y0);

	mpf_clear(tmp);
	mpf_clear(tmp1);

	return times;
}

/*                                                 */
/* Implicit Runge-Kutta Method with fixed stepsize */
/*                                                 */
void mpf_irk_fs(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, long int div_num, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), void(* jfunc)(MPFMatrix ymat, mpf_t x0, MPFVector y0), unsigned long coef_type, mpf_t rtol, mpf_t atol, long int maxtimes)
{
	long int steps, i, itimes;
	MPFVector old_y, tmp_y, lf_tmp;
	mpf_t tmp[3], new_x, old_x, in_h;
	MPFVector coef_c, coef_w;
	MPFMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case IRK_GAUSS_1_2:
			coef_c = init2_mpfvector(1, y->prec);
			coef_w = init2_mpfvector(1, y->prec);
			coef_a = init2_mpfmatrix(1, 1, y->prec);
			steps = 1;
			break;
		
		case IRK_GAUSS_2_4:
			coef_c = init2_mpfvector(2, y->prec);
			coef_w = init2_mpfvector(2, y->prec);
			coef_a = init2_mpfmatrix(2, 2, y->prec);
			steps = 2;
			break;

		case IRK_GAUSS_3_6:
			coef_c = init2_mpfvector(3, y->prec);
			coef_w = init2_mpfvector(3, y->prec);
			coef_a = init2_mpfmatrix(3, 3, y->prec);
			steps = 3;
			break;

		default:
			fprintf(stderr, "mpf_irk_ex: Such a coefficient set not exist.\n");
			return;
	}

	get_mpf_irk_coef(coef_type, y->prec, coef_c, coef_a, coef_w);

	init_mpf_irk_1step(y->dim, y->prec, steps);

//	printf("c: \n"); print_mpfvector(coef_c);
//	printf("w: \n"); print_mpfvector(coef_w);
//	printf("a: \n"); print_mpfmatrix(coef_a);

	/* init */
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));

	/* set */
	mpf_set(old_x, x0);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_h, x, x0);
	if(div_num <= 0)
	{
		fprintf(stderr, "mpf_irk_fs: Number of division is illegal.\n");
		return ;
	}
	mpf_div_ui(in_h, in_h, (unsigned long)div_num);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf("times\n");
	}

	/* main loop */
	for(steps = 0; steps < div_num; steps++)
	{

		/* tmp_y := y0 + h * irk(x0, y0) */
		itimes = mpf_irk_1step(tmp_y, old_x, old_y, in_h, func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, tmp_y);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", mpf2double(old_x));
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", mpf2double(gmpfvi(old_y, i)));
			printf(" %ld", itimes);
			printf("\n");
		}

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	/* output */
	printf("OK! mpf_irk_fs has been finished.\n");
	printf("Stepsize  : ");
		 mpf_out_str(stdout, 10, 0, in_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");

	/* clear */
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	free_mpfvector(coef_c);
	free_mpfvector(coef_w);
	free_mpfmatrix(coef_a);

}

/*                                                                       */
/* Implicit Runge-Kutta Method with stepsize control using extrapolation */
/*                                                                       */
void mpf_irk_ex(FILE *fp, mpf_t x, MPFVector y, mpf_t x0, MPFVector y0, mpf_t max_h, void(* func)(MPFVector y, mpf_t x0, MPFVector y0), void(* jfunc)(MPFMatrix ymat, mpf_t x0, MPFVector y0), unsigned long coef_type, mpf_t rtol, mpf_t atol, long int maxtimes)
{
	long int steps, i, itimes[2];
	MPFVector new_y[2], old_y, tmp_y, lf_tmp;
	mpf_t in_max_h, tmp[3], new_x, old_x, in_h, min_h;
	MPFVector coef_c, coef_w;
	MPFMatrix coef_a;

	/* set coef */
	switch(coef_type)
	{
		case IRK_GAUSS_1_2:
			coef_c = init2_mpfvector(1, y->prec);
			coef_w = init2_mpfvector(1, y->prec);
			coef_a = init2_mpfmatrix(1, 1, y->prec);
			steps = 1;
			break;
		
		case IRK_GAUSS_2_4:
			coef_c = init2_mpfvector(2, y->prec);
			coef_w = init2_mpfvector(2, y->prec);
			coef_a = init2_mpfmatrix(2, 2, y->prec);
			steps = 2;
			break;

		case IRK_GAUSS_3_6:
			coef_c = init2_mpfvector(3, y->prec);
			coef_w = init2_mpfvector(3, y->prec);
			coef_a = init2_mpfmatrix(3, 3, y->prec);
			steps = 3;
			break;

		default:
			fprintf(stderr, "mpf_irk_ex: Such a coefficient set not exist.\n");
			return;
	}

	get_mpf_irk_coef(coef_type, y->prec, coef_c, coef_a, coef_w);

	init_mpf_irk_1step(y->dim, y->prec, steps);

//	printf("c: \n"); print_mpfvector(coef_c);
//	printf("w: \n"); print_mpfvector(coef_w);
//	printf("a: \n"); print_mpfmatrix(coef_a);

	/* init */
	new_y[0] = init2_mpfvector(y->dim, y->prec);
	new_y[1] = init2_mpfvector(y->dim, y->prec);
	old_y = init2_mpfvector(y->dim, y->prec);
	tmp_y = init2_mpfvector(y->dim, y->prec);
	lf_tmp = init2_mpfvector(y->dim, y->prec);
	mpf_init2(new_x, mpf_get_prec(x));
	mpf_init2(old_x, mpf_get_prec(x));
	mpf_init2(tmp[0], mpf_get_prec(x));
	mpf_init2(tmp[1], mpf_get_prec(x));
	mpf_init2(tmp[2], mpf_get_prec(x));
	mpf_init2(in_h, mpf_get_prec(x));
	mpf_init2(in_max_h, mpf_get_prec(x));
	mpf_init2(min_h, mpf_get_prec(x));

	/* set */
	mpf_set(old_x, x0);
	mpf_set(min_h, max_h);
	subst_mpfvector(old_y, y0);

	/* check interval */
	/* in_max_h := min((x - x0) / 2, max_h) */
	mpf_sub(in_max_h, x, x0);
	mpf_div_ui(in_max_h, in_max_h, 2UL);
	if(mpf_cmp(max_h, in_max_h) < 0)
		mpf_set(in_max_h, max_h);

	/* output */
	if(fp != NULL)
	{
		printf("           x             ");
		for(i = 0; i < y->dim; i++)
			printf("         y[%5d]          ", i);
		printf(" stepsize   ");
		printf("times(h, h/2)\n");
	}

	/* main loop */
	for(steps = 0; ; steps++)
	{

		/* set default interval */
		mpf_set(in_h, in_max_h);

		/* calc */

		/* tmp_y := y0 + h * irk(x0, y0) */
		mpf_irk_1step(tmp_y, old_x, old_y, in_h, func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);
		while(1)
		{
	
			/* h := h / 2 */
			mpf_div_ui(tmp[2], in_h, 2UL);

			/* new_y[0] := y0 + (h/2) * irk(x0, y0) */
			/* new_y[1] := new_y[0] + (h/2) * irk(x0 + h/2, new_y[0]) */
			itimes[0] = mpf_irk_1step(new_y[0], old_x, old_y, tmp[2], func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);

			mpf_add(tmp[0], old_x, tmp[2]);
			itimes[1] = mpf_irk_1step(new_y[1], tmp[0], new_y[0], tmp[2], func, jfunc, coef_c, coef_a, coef_w, atol, rtol, maxtimes);

			/* calc error */
			sub_mpfvector(lf_tmp, new_y[1], tmp_y);
			add2_mpfvector(new_y[1], lf_tmp);

			/* || new_y[1] - tmp_y || <= || new_y[1] || * rtol + atol */
			normi_mpfvector(tmp[0], lf_tmp);
			normi_mpfvector(tmp[1], new_y[1]);
			mpf_mul(tmp[1], tmp[1], rtol);
			mpf_add(tmp[1], tmp[1], atol);
			if(mpf_cmp(tmp[0], tmp[1]) <= 0)
					break;

			/* check in_h */
			if(mpf_cmp_si(tmp[2], 0) <= 0)
			{
				fprintf(stderr, "mpf_irk: inner stepsize is too small.\n");
				break;
			}

			mpf_set(in_h, tmp[2]);
			subst_mpfvector(tmp_y, new_y[0]);

//			printf("%10.7e %25.17e %25.17e\n", mpf2double(in_h), mpf2double(gmpfvi(lf_tmp, 0)), mpf2double(gmpfvi(tmp_y, 0)), mpf2double(gmpfvi(new_y[1], 0)));
		}

		if(mpf_cmp(min_h, in_h) > 0)
			mpf_set(min_h, in_h);

		/* set new x, y */
		mpf_add(old_x, old_x, in_h);
		subst_mpfvector(old_y, new_y[1]);

		/* output */
		if(fp != NULL)
		{
			printf("%25.17e ", mpf2double(old_x));
			for(i = 0; i < old_y->dim; i++)
				printf("%25.17e ", mpf2double(gmpfvi(old_y, i)));
			printf("%15.7e ", mpf2double(in_h));
			printf(" %ld %ld", itimes[0], itimes[1]);
			printf("\n");
		}

		if(mpf_cmp(old_x, x) >= 0)
			break;

	}

	/* finish! */
	subst_mpfvector(y, old_y);

	/* output */
	printf("OK! mpf_irk_ex has been finished.\n");
	printf("Minimum/Maxmum stepsize  : ");
		 mpf_out_str(stdout, 10, 0, min_h);
		 printf(" / ");
		 mpf_out_str(stdout, 10, 0, max_h);
		 printf("\n");
	printf("Number of steps  : %ld\n", steps);
	printf("Integral interval : ");
		 printf("[ ");
		 mpf_out_str(stdout, 10, 0, x0);
		 printf(", ");
		 mpf_out_str(stdout, 10, 0, old_x);
		 printf(" ]\n");
	printf("Numerical solution: \n");
		 print_mpfvector(y);
		 printf("\n");

	/* clear */
	free_mpfvector(new_y[0]);
	free_mpfvector(new_y[1]);
	free_mpfvector(old_y);
	free_mpfvector(tmp_y);
	free_mpfvector(lf_tmp);
	mpf_clear(new_x);
	mpf_clear(old_x);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);
	mpf_clear(in_h);
	mpf_clear(in_max_h);
	mpf_clear(min_h);
	free_mpfvector(coef_c);
	free_mpfvector(coef_w);
	free_mpfmatrix(coef_a);

}

#endif
