/***********************************************/
/* Algebraic Solvers for                       */
/*     2, 3 and 4th Degree Algebraic Equations */
/*                                             */
/* Copyright (C) 2007- Tomonori Kouya          */
/*                                             */
/* This library is free software; you can      */
/* redistribute it and/or modify it under the  */
/* terms of the GNU Lesser General Public      */
/* License as published by the Free Software   */
/* Foundation; either version 2.1 of the       */
/* License, or (at your option) any later      */
/* version.                                    */
/*                                             */
/* This library is distributed in the hope     */
/* that it will be useful, but WITHOUT ANY     */
/* WARRANTY; without even the implied warranty */
/* of MERCHANTABILITY or FITNESS FOR A         */
/* PARTICULAR PURPOSE.  See the GNU Lesser     */
/* General Public License for more details.    */
/*                                             */
/* You should have received a copy of the GNU  */
/* Lesser General Public License along with    */
/* this library; if not, write to the Free     */
/* Software Foundation, Inc., 51 Franklin      */
/* Street, Fifth Floor, Boston, MA  02110-1301 */
/* USA                                         */
/***********************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

/***********************************************/
/* Solver of Quadratic Equation                */
/***********************************************/

/* double */

/* coef[2] x^2 + coef[1] x + coef[0] = 0 */
/* Output: ans_re[0] + sqrt(-1)*ans_im[0], ans_re[1] + sqrt(-1) * ans_im[1] */
int dquadratic_eq(double ans_re[2], double ans_im[2], double coef[3])
{
	double d, den, tmp;

//	printf("coef: %10.3e x^2 + %10.3e x + %10.3e = 0\n", coef[2], coef[1], coef[0]);

	ans_re[0] = 0.0;
	ans_re[1] = 0.0;
	ans_im[0] = 0.0;
	ans_im[1] = 0.0;

	if(coef[2] == 0.0)
	{
		if(coef[1] == 0.0)
		{
			fprintf(stderr, "ERROR: No answers (dquadraric_eq)\n");
			return -1;
		}
		ans_re[0] = -coef[0] / coef[1];
		return 1; // number of answers
	}

	den = 2.0 * coef[2];
	d = coef[1] * coef[1] - 4.0 * coef[2] * coef[0];
	ans_re[0] = -coef[1] / den;

	/* anss are real numbers */
	if(d == 0.0)
	{
		ans_re[1] = ans_re[0];
	}
	else if(d > 0)
	{
		tmp = sqrt(d);
		if(coef[1] > 0.0)
			tmp = -tmp / den;
		else
			tmp = tmp / den;
		ans_re[0] -= tmp;
		ans_re[1] = coef[0] / (coef[2] * ans_re[0]);
	}
	/* complex numbers */
	else
	{
		tmp = sqrt(-d) / den;
		ans_re[1] = ans_re[0];
		ans_im[0] = tmp;
		ans_im[1] = -tmp;
	}
	return 2;
}

/* MPF */

#ifdef USE_GMP
/* coef[2] x^2 + coef[1] x + coef[0] = 0 */
/* Output: ans_re[0] + sqrt(-1)*ans_im[0], ans_re[1] + sqrt(-1) * ans_im[1] */
int mpfquadratic_eq(mpf_t ans_re[2], mpf_t ans_im[2], mpf_t coef[3])
{
	unsigned prec;
	mpf_t d, den, tmp;

//	printf("coef: %10.3e x^2 + %10.3e x + %10.3e = 0\n", coef[2], coef[1], coef[0]);

	prec = (mpf_get_prec(ans_re[0]) > mpf_get_prec(ans_re[1])) ? mpf_get_prec(ans_re[0]) : mpf_get_prec(ans_re[1]);

	mpf_set_ui(ans_re[0], 0UL);
	mpf_set_ui(ans_re[1], 0UL);
	mpf_set_ui(ans_im[0], 0UL);
	mpf_set_ui(ans_im[1], 0UL);

	if(mpf_cmp_ui(coef[2], 0UL) == 0)
	{
		if(mpf_cmp_ui(coef[1], 0UL) == 0)
		{
			fprintf(stderr, "ERROR: No answers (mpfquadraric_eq)\n");
			return -1;
		}
		mpf_neg(ans_re[0], coef[0]);
		mpf_div(ans_re[0], ans_re[0], coef[1]);
		return 1; // number of answers
	}

	/* Initialize */
	mpf_init2(d, prec);
	mpf_init2(den, prec);
	mpf_init2(tmp, prec);

	mpf_mul_ui(den, coef[2], 2UL);
	/* d = coef[1] * coef[1] - 4.0 * coef[2] * coef[0]; */
	mpf_mul(d, coef[1], coef[1]);
	mpf_mul(tmp, coef[2], coef[0]);
	mpf_mul_ui(tmp, tmp, 4UL);
	mpf_sub(d, d, tmp);

	/* ans_re[0] = -coef[1] / den; */
	mpf_neg(ans_re[0], coef[1]);
	mpf_div(ans_re[0], ans_re[0], den);

	/* anss are real numbers */
	if(mpf_cmp_ui(d, 0UL) == 0)
	{
		mpf_set(ans_re[1], ans_re[0]);
	}
	else if(mpf_cmp_ui(d, 0UL) > 0)
	{
		mpf_sqrt(tmp, d);
		if(mpf_cmp_ui(coef[1], 0UL) > 0)
		{
			mpf_neg(tmp, tmp);
			mpf_div(tmp, tmp, den);
		}
		else
			mpf_div(tmp, tmp, den);
		mpf_sub(ans_re[0], ans_re[0], tmp);
		mpf_mul(tmp, coef[2], ans_re[0]);
		mpf_div(ans_re[1], coef[0], tmp);
	}
	/* complex numbers */
	else
	{
		/* tmp = sqrt(-d) / den; */
		mpf_neg(tmp, d);
		mpf_sqrt(tmp, tmp);
		mpf_div(tmp, tmp, den);

		mpf_set(ans_re[1], ans_re[0]);
		mpf_set(ans_im[0], tmp);
		mpf_neg(ans_im[1], tmp);
	}

	/* free */
	mpf_clear(d);
	mpf_clear(den);
	mpf_clear(tmp);

	return 2;
}
#endif
/***********************************************/
/* End of algebraic_eq.c                       */
/***********************************************/

