/**********************************************/
/* fread_write.c: Read and Write files        */
/*             to store *matrix, *vector etc. */
/*                                            */
/* Copyright (c) 2005- Tomonori Kouya         */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
#include <stdio.h>
#include <stdlib.h>

#include "bnc.h"

void fread_dmatrix(FILE *fp, DMatrix mat)
{
	long int i, row_index, col_index;
	double tmp;

	for(i = 0; i < mat->row_dim * mat->col_dim; i++)
	{
		fscanf(fp, "%ld, %ld, %lf", &row_index, &col_index, &tmp);
		set_dmatrix_ij(mat, row_index, col_index, tmp);
		if(feof(fp) != 0)
			break;
	}
}

void fread_dmatrix_fname(const char *fname, DMatrix mat)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fread_dmatrix(fp, mat);

	fclose(fp);
}

void fwrite_dmatrix(FILE *fp, DMatrix mat)
{
	long int i, j;

	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->col_dim; j++)
			fprintf(fp, "%ld, %ld, %lf\n", i, j, get_dmatrix_ij(mat, i, j));
	}
}

void fwrite_dmatrix_fname(const char *fname, DMatrix mat)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "w");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fwrite_dmatrix(fp, mat);

	fclose(fp);
}

void fread_dpolycoef(FILE *fp, DPoly p, long int maxdeg)
{
	long int i, index;
	double tmp;

	for(i = 0; i <= maxdeg; i++)
	{
		fscanf(fp, "%ld, %lf", &index, &tmp);
		set_dpoly_i(p, index, tmp);
	}
}

void fread_dpolycoef_fname(const char *fname, DPoly p, long int maxdeg)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fread_dpolycoef(fp, p, maxdeg);

	fclose(fp);
}

void fread_dvector(FILE *fp, DVector vec)
{
	long int i, index;
	double tmp;

	for(i = 0; i < vec->dim; i++)
	{
		fscanf(fp, "%ld, %lf", &index, &tmp);
		set_dvector_i(vec, index, tmp);
	}
}

void fread_dvector_fname(const char *fname, DVector vec)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fread_dvector(fp, vec);

	fclose(fp);
}

void fwrite_dvector(FILE *fp, DVector vec)
{
	long int i;

	for(i = 0; i < vec->dim; i++)
		fprintf(fp, "%d, %25.17e\n", i, get_dvector_i(vec, i));
}

void fwrite_dvector_fname(const char *fname, DVector vec)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "w");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fwrite_dvector(fp, vec);

	fclose(fp);
}

#ifdef USE_GMP
void fread_mpfmatrix(FILE *fp, MPFMatrix mat)
{
	long int i, row_index, col_index;
	char tmpsrc[65536]; /* maximum degits */
	mpf_t tmp;

	mpf_init2(tmp, mat->prec);

	for(i = 0; i < mat->row_dim * mat->col_dim; i++)
	{
		fscanf(fp, "%ld, %ld, %s", &row_index, &col_index, tmpsrc);
		mpf_set_str(tmp, tmpsrc, 10);
		set_mpfmatrix_ij(mat, row_index, col_index, tmp);
		if(feof(fp) != 0)
			break;
	}

	mpf_clear(tmp);
}

void fread_mpfmatrix_fname(const char *fname, MPFMatrix mat)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fread_mpfmatrix(fp, mat);

	fclose(fp);
}

void fwrite_mpfmatrix(FILE *fp, MPFMatrix mat)
{
	long int i, j;
	char tmpsrc[65536]; /* maximum degits */
	mpf_t tmp;

	mpf_init2(tmp, mat->prec);

	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->col_dim; j++)
		{
			if(mpf_cmp_ui(get_mpfmatrix_ij(mat, i, j), 0UL) != 0)
			{
				fprintf(fp, "%ld, %ld, ", i, j);
				mpf_out_str(fp, 10, 0, get_mpfmatrix_ij(mat, i, j));
				fprintf(fp, "\n");
			}
		}
	}

	mpf_clear(tmp);
}

void fwrite_mpfmatrix_fname(const char *fname, MPFMatrix mat)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "w");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fwrite_mpfmatrix(fp, mat);

	fclose(fp);
}

void fread_mpfpolycoef(FILE *fp, MPFPoly p, long int maxdeg)
{
	long int i, index;
	char tmpsrc[65536]; /* maximum degits */
	mpf_t tmp;

	mpf_init2(tmp, p->prec);

	for(i = 0; i <= maxdeg; i++)
	{
		fscanf(fp, "%ld, %s", &index, tmpsrc);
//		printf("%d: %s\n", index, tmpsrc);
		mpf_set_str(tmp, tmpsrc, 10);
		set_mpfpoly_i(p, index, tmp);
	}

	mpf_clear(tmp);
}

void fread_mpfpolycoef_fname(const char *fname, MPFPoly p, long int maxdeg)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fread_mpfpolycoef(fp, p, maxdeg);

	fclose(fp);
}

void fread_mpfvector(FILE *fp, MPFVector vec)
{
	long int i, index;
	char tmpsrc[65536]; /* maximum degits */
	mpf_t tmp;

	mpf_init2(tmp, vec->prec);

	for(i = 0; i < vec->dim; i++)
	{
		fscanf(fp, "%ld, %s", &index, tmpsrc);
		mpf_set_str(tmp, tmpsrc, 10);
		set_mpfvector_i(vec, index, tmp);
	}

	mpf_clear(tmp);
}

void fread_mpfvector_fname(const char *fname, MPFVector vec)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fread_mpfvector(fp, vec);

	fclose(fp);
}

void fwrite_mpfvector(FILE *fp, MPFVector vec)
{
	long int i;

	for(i = 0; i < vec->dim; i++)
	{
		fprintf(fp, "%ld, ", i);
		mpf_out_str(fp, 10, 0, get_mpfvector_i(vec, i));
		fprintf(fp, "\n");
	}
}

void fwrite_mpfvector_fname(const char *fname, MPFVector vec)
{
	FILE *fp;

	if(fname == NULL)
	{
		fprintf(stderr, "file name is undefined!\n");
		return;
	}

	fp = fopen(fname, "w");
	if(fp == NULL)
	{
		fprintf(stderr, "%s cannot be opened!\n", fname);
		return;
	}

	fwrite_mpfvector(fp, vec);

	fclose(fp);
}

#endif
