/**********************************************/
/* linear.c: Vector, Matrix                   */
/* Copyright (c) 2000- Tomonori Kouya         */
/*                                            */
/* Version 0.1, 2005.07/11:                   */
/*                     append copy_*vector_ij */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

/*************************************************/
/* Functions for Vector Types                    */
/*                                               */
/* Initialize:                                   */
/*   FVector init_fvector(long int dimension)    */
/*   DVector init_dvector(long int dimension)    */
/*   MPFVector init_mpfvector(long int dimension)*/
/*   MPFVector init2_mpfvector(long int dimension, unsigned long mbits)*/
/* Free:                                         */
/*   void free_fvector(FVector vec)              */
/*   void free_dvector(DVector vec)              */
/*   void free_mpfvector(MPFVector vec)          */
/* Get & Set Values:                             */
/*   float get_fvector_i(FVector vec, long int index) */
/*   double get_dvector_i(DVector vec, long int index) */
/*   mpf_t *get_mpfvector_i(MPFVector vec, long int index) */
/*   void set_fvector_i(FVector vec, long int index, float val) */
/*   void set_dvector_i(DVector vec, long int index, double val) */
/*   void set_mpfvector_i(MPFVector vec, long int index, mpf_t val) */
/*   void set_mpfvector_i_d(MPFVector vec, long int index, double val) */
/* Output:                                       */
/*   void print_fvector(FVector vec)             */
/*   void print_dvector(DVector vec)             */
/*   void print_mpfvector(MPFVector vec)         */
/*   void print_fdmpfvector(FVector fv, DVector dv, MPFVector mpfv) */
/*************************************************/
FVector init_fvector(long int dimension)
{
	FVector ret = NULL;
	long int i;

	if(dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_fvector\n");
		return ret;
	}

	ret = (FVector)malloc(sizeof(fvector));
	if(ret == NULL)
		return ret;

	ret->element = (float *)calloc(sizeof(float), dimension);
	if(ret->element == NULL)
		return ret;

	/* All 0 */
	for(i = 0; i < dimension; i++)
		*(ret->element + i) = 0.0;

	ret->dim = dimension;

	return ret;
}

DVector init_dvector(long int dimension)
{
	DVector ret = NULL;
	long int i;

	if(dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_dvector\n");
		return ret;
	}

	ret = (DVector)malloc(sizeof(dvector));
	if(ret == NULL)
		return ret;

	ret->element = (double *)calloc(sizeof(double), dimension);
	if(ret->element == NULL)
		return ret;

	/* All 0 */
	for(i = 0; i < dimension; i++)
		*(ret->element + i) = 0.0;

	ret->dim = dimension;

	return ret;
}

#ifdef USE_GMP
MPFVector init_mpfvector(long int dimension)
{
	MPFVector ret = NULL;
	long int i;

	if(dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_mpfvector\n");
		return ret;
	}

	ret = (MPFVector)malloc(sizeof(mpfvector));
	if(ret == NULL)
		return ret;

	ret->element = (mpf_t *)calloc(sizeof(mpf_t), dimension);
//	ret->element = (mpf_t *)malloc(sizeof(mpf_t) * dimension);
	if(ret->element == NULL)
		return NULL;

	/* All 0 */
	for(i = 0; i < dimension; i++)
	{
		mpf_init((mpf_ptr)(ret->element + i));
		mpf_set_ui((mpf_ptr)(ret->element + i), 0UL);
		if((ret->element+i) == NULL)
			return NULL;
	}

	ret->dim = dimension;

	ret->prec = get_bnc_default_prec();

	return ret;
}

/* mbits ... A number of at least bits of mantissa */
MPFVector init2_mpfvector(long int dimension, unsigned long int mbits)
{
	MPFVector ret = NULL;
	long int i;

	if(dimension <= 0)
	{
		fprintf(stderr, "ERROR: init2_mpfvector\n");
		return ret;
	}

	ret = (MPFVector)malloc(sizeof(mpfvector));
	if(ret == NULL)
		return ret;

	ret->element = (mpf_t *)calloc(sizeof(mpf_t), dimension);
//	ret->element = (mpf_t *)malloc(sizeof(mpf_t) * dimension);
	if(ret->element == NULL)
		return NULL;

	/* All 0 */
	for(i = 0; i < dimension; i++)
	{
		mpf_init2((mpf_ptr)(ret->element + i), mbits);
		mpf_set_ui((mpf_ptr)(ret->element + i), 0UL);
		if((ret->element+i) == NULL)
			return NULL;
	}

	ret->dim = dimension;

	ret->prec = mbits;

	return ret;
}
#endif

void free_fvector(FVector vec)
{
	if(vec == NULL)
		return;

	if(vec->element != NULL)
		free(vec->element);

	free(vec);
}
void free_dvector(DVector vec)
{
	if(vec == NULL)
		return;

	if(vec->element != NULL)
		free(vec->element);

	free(vec);
}

#ifdef USE_GMP
void free_mpfvector(MPFVector vec)
{
	long int i;

	if(vec == NULL)
		return;

	if(vec->element != NULL)
	{
		for(i = 0; i < vec->dim; i++)
			mpf_clear((mpf_ptr)(vec->element + i));
	}

//	free(&(vec->dim));
//	free(&(vec->prec));
	free(vec);

}
#endif

float get_fvector_i(FVector vec, long int index)
{
	return *(vec->element + index);
}

double get_dvector_i(DVector vec, long int index)
{
	return *(vec->element + index);
}

#ifdef USE_GMP
mpf_ptr get_mpfvector_i(MPFVector vec, long int index)
{
	return *(vec->element + index);
}
#endif

void set_fvector_i(FVector vec, long int index, float val)
{
	*(vec->element + index) = val;
}

void set_dvector_i(DVector vec, long int index, double val)
{
	*(vec->element + index) = val;
}

#ifdef USE_GMP
void set_mpfvector_i(MPFVector vec, long int index, mpf_t val)
{
	mpf_set((mpf_ptr)(vec->element + index), val);
}

void set_mpfvector_i_d(MPFVector vec, long int index, double val)
{
	mpf_set_d(*(vec->element + index), val);
}

void set_mpfvector_i_str(MPFVector vec, long int index, const char *str, int base)
{
	mpf_set_str(*(vec->element + index), str, base);
}

void set_mpfvector_i_ui(MPFVector vec, long int index, unsigned long val)
{
	mpf_set_ui(*(vec->element + index), val);
}

/* get precision of MPFVector */
unsigned long int prec_mpfvector(MPFVector vec)
{
	return vec->prec;
}

/* search minimam precision in MPFVector */
unsigned long int minprec_mpfvector(MPFVector vec)
{
	unsigned long int prec, tmp;
	long int i;

	prec = mpf_get_prec(get_mpfvector_i(vec, 0));
	for(i = 1; i < vec->dim; i++)
	{
		tmp = mpf_get_prec(get_mpfvector_i(vec, i));
		if(prec > tmp)
			prec = tmp;
	}

	return prec;
}

/* search maximam precision in MPFVector */
unsigned long int maxprec_mpfvector(MPFVector vec)
{
	unsigned long int prec, tmp;
	long int i;

	prec = mpf_get_prec(get_mpfvector_i(vec, 0));
	for(i = 1; i < vec->dim; i++)
	{
		tmp = mpf_get_prec(get_mpfvector_i(vec, i));
		if(prec < tmp)
			prec = tmp;
	}

	return prec;
}
#endif

void print_fvector(FVector fv)
{
	long int i;

	for(i = 0; i < fv->dim; i++)
		printf("%5ld %15.7e\n", i, get_fvector_i(fv, i));
}

void print_dvector(DVector dv)
{
	long int i, dim;

	for(i = 0; i < dv->dim; i++)
		printf("%5ld %25.17e\n", i, get_dvector_i(dv, i));
}

void print_fdvector(FVector fv, DVector dv)
{
	long int i, dim;

	dim = fv->dim;
	if(dim > dv->dim)
		dim = dv->dim;

	for(i = 0; i < dim; i++)
		printf("%5ld %15.7e %25.17e\n", i, get_fvector_i(fv, i), get_dvector_i(dv, i));
}

#ifdef USE_GMP
void print_mpfvector(MPFVector vec)
{
	long int i;

	for(i = 0; i < vec->dim; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(vec, i));
		printf("\n");
	}
}

void print_fdmpfvector(FVector fv, DVector dv, MPFVector mpfv)
{
	long int i;

	for(i = 0; i < fv->dim; i++)
	{
		printf("%5ld %15.7e %25.17e ", i, get_fvector_i(fv, i), get_dvector_i(dv, i));
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfv, i));
		printf("\n");
	}
}
#endif

/*************************************************/
/* Function for Matrix Types                     */
/*                                               */
/* Initialize:                                   */
/*   FMatrix init_fmatrix(long int row_dimension, long int col_dimension)    */
/*   DMatrix init_dmatrix(long int row_dimension, long int col_dimension)    */
/*   MPFMatrix init_mpfmatrix(long int row_dimension, long int col_dimension)*/
/*   MPFMatrix init2_mpfmatrix(long int row_dimension, long int col_dimension, unsigned long mbits) */
/* Free:                                         */
/*   void free_fmatrix(FMatrix mat)              */
/*   void free_dmatrix(DMatrix mat)              */
/*   void free_mpfmatrix(MPFMatrix mat)          */
/* Get & Set:
/*   float get_fmatrix_ij(FMatrix mat, long int row_index, long int col_index) */
/*   float get_dmatrix_ij(DMatrix mat, long int row_index, long int col_index) */
/*   float get_mpfmatrix_ij(MPFMatrix mat, long int row_index, long int col_index) */
/*   void set_fmatrix_ij(FMatrix mat, long int row_index, long int col_index, float val) */
/*   void set_dmatrix_ij(DMatrix mat, long int row_index, long int col_index, double val) */
/*   void set_mpfmatrix_ij(MPFMatrix mat, long int row_index, long int col_index, mpf_t val) */
/*   void set_mpfmatrix_ij_d(MPFMatrix mat, long int row_index, long int col_index, double val) */
/* Output:                                       */
/*   void print_fmatrix(FMatrix mat)             */
/*   void print_dmatrix(DMatrix mat)             */
/*   void print_mpfmatrix(MPFMatrix mat)         */
/*************************************************/
FMatrix init_fmatrix(long int row_dimension, long int col_dimension)
{
	FMatrix ret = NULL;
	long int i, j;

	/* Check Dimentions */
	if(row_dimension <= 0 || col_dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_fmatrix\n");
		return ret;
	}

	ret = (FMatrix)malloc(sizeof(fmatrix));
	if(ret == NULL)
	{
		fprintf(stderr, "ERROR: init_fmatrix\n");
		return ret;
	}

	ret->element = (float *)calloc(sizeof(float), row_dimension * col_dimension);
	if(ret->element == NULL)
	{
		fprintf(stderr, "ERROR: init_fmatrix\n");
		return ret;
	}

	/* All 0 */
	for(i = 0; i < row_dimension; i++)
		for(j = 0; j < col_dimension; j++)
			*(ret->element + i * col_dimension + j) = (float)0.0;

	ret->row_dim = row_dimension;
	ret->col_dim = col_dimension;

	return ret;
}

DMatrix init_dmatrix(long int row_dimension, long int col_dimension)
{
	DMatrix ret = NULL;
	long int i, j;

	if(row_dimension <= 0 || col_dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_dmatrix\n");
		return ret;
	}

	ret = (DMatrix)malloc(sizeof(dmatrix));
	if(ret == NULL)
		return ret;

	ret->element = (double *)calloc(sizeof(double), row_dimension * col_dimension);
	if(ret->element == NULL)
		return ret;

	/* All 0 */
	for(i = 0; i < row_dimension; i++)
		for(j = 0; j < col_dimension; j++)
			*(ret->element + i * col_dimension + j) = (double)0.0;

	ret->row_dim = row_dimension;
	ret->col_dim = col_dimension;

	return ret;
}

#ifdef USE_GMP
MPFMatrix init_mpfmatrix(long int row_dimension, long int col_dimension)
{
	MPFMatrix ret = NULL;
	long int i, j;

	if(row_dimension <= 0 || col_dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_mpfmatrix\n");
		return ret;
	}

	ret = (MPFMatrix)malloc(sizeof(mpfmatrix));
	if(ret == NULL)
		return ret;

	ret->element = (mpf_t *)calloc(sizeof(mpf_t), row_dimension * col_dimension);
	if(ret->element == NULL)
		return ret;

	/* All 0 */
	for(i = 0; i < row_dimension; i++)
		for(j = 0; j < col_dimension; j++)
		{
			mpf_init(*(ret->element + i * col_dimension + j));
			mpf_set_ui(*(ret->element + i * col_dimension + j), 0UL);
		}

	ret->row_dim = row_dimension;
	ret->col_dim = col_dimension;

	ret->prec = get_bnc_default_prec();

	return ret;
}
/* mbits ... A number of at least bits of mantissa */
MPFMatrix init2_mpfmatrix(long int row_dimension, long int col_dimension, unsigned long mbits)
{
	MPFMatrix ret = NULL;
	long int i, j;

	if(row_dimension <= 0 || col_dimension <= 0)
	{
		fprintf(stderr, "ERROR: init_mpfmatrix\n");
		return ret;
	}

	ret = (MPFMatrix)malloc(sizeof(mpfmatrix));
	if(ret == NULL)
		return ret;

	ret->element = (mpf_t *)calloc(sizeof(mpf_t), row_dimension * col_dimension);
	if(ret->element == NULL)
		return ret;

	/* All 0 */
	for(i = 0; i < row_dimension; i++)
		for(j = 0; j < col_dimension; j++)
		{
			mpf_init2(*(ret->element + i * col_dimension + j), mbits);
			mpf_set_ui(*(ret->element + i * col_dimension + j), 0UL);
		}

	ret->row_dim = row_dimension;
	ret->col_dim = col_dimension;

	ret->prec = mbits;

	return ret;
}
#endif

void free_fmatrix(FMatrix mat)
{
	if(mat == NULL)
		return;

	if(mat->element != NULL)
		free(mat->element);

	free(mat);
}

void free_dmatrix(DMatrix mat)
{
	if(mat == NULL)
		return;

	if(mat->element != NULL)
		free(mat->element);

	free(mat);
}

#ifdef USE_GMP
void free_mpfmatrix(MPFMatrix mat)
{
	long int i, j;

	if(mat == NULL)
		return;

	if(mat->element != NULL)
		for(i = 0; i < mat->row_dim; i++)
			for(j = 0; j < mat->col_dim; j++)
				mpf_clear(*(mat->element + i * mat->col_dim + j));
	free(mat);
}
#endif

float get_fmatrix_ij(FMatrix mat, long int row_index, long int col_index)
{
	return (float)(*(mat->element + row_index * mat->col_dim + col_index));
}

double get_dmatrix_ij(DMatrix mat, long int row_index, long int col_index)
{
	return *(mat->element + row_index * mat->col_dim + col_index);
}

#ifdef USE_GMP
mpf_ptr get_mpfmatrix_ij(MPFMatrix mat, long int row_index, long int col_index)
{
	return *(mat->element + row_index * mat->col_dim + col_index);
}
#endif

void set_fmatrix_ij(FMatrix mat, long int row_index, long int col_index, float val)
{
	*(mat->element + row_index * mat->col_dim + col_index) = (float)val;
}

void set_dmatrix_ij(DMatrix mat, long int row_index, long int col_index, double val)
{
	*(mat->element + row_index * mat->col_dim + col_index) = val;
}

#ifdef USE_GMP
void set_mpfmatrix_ij(MPFMatrix mat, long int row_index, long int col_index, mpf_t val)
{
	mpf_set(*(mat->element + row_index * mat->col_dim + col_index), val);
}

void set_mpfmatrix_ij_d(MPFMatrix mat, long int row_index, long int col_index, double val)
{
	mpf_set_d((mpf_ptr)(mat->element + row_index * mat->col_dim + col_index), val);
}

void set_mpfmatrix_ij_str(MPFMatrix mat, long int row_index, long int col_index, const char *str, int base)
{
	mpf_set_str((mpf_ptr)(mat->element + row_index * mat->col_dim + col_index), str, base);
}

void set_mpfmatrix_ij_ui(MPFMatrix mat, long int row_index, long int col_index, unsigned long val)
{
	mpf_set_d((mpf_ptr)(mat->element + row_index * mat->col_dim + col_index), val);
}

/* get precision of MPFMatrix */
unsigned long int prec_mpfmatrix(MPFMatrix mat)
{
	return mat->prec;
}

/* search minimam precision in MPFMatrix */
unsigned long int minprec_mpfmatrix(MPFMatrix mat)
{
	unsigned long int prec, tmp;
	long int i, j;

	prec = mpf_get_prec(get_mpfmatrix_ij(mat, 0, 0));
	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->col_dim; j++)
		{
			tmp = mpf_get_prec(get_mpfmatrix_ij(mat, i, j));
			if(prec > tmp)
				prec = tmp;
		}
	}

	return prec;
}

/* search maximam precision in MPFMatrix */
unsigned long int maxprec_mpfmatrix(MPFMatrix mat)
{
	unsigned long int prec, tmp;
	long int i, j;

	prec = mpf_get_prec(get_mpfmatrix_ij(mat, 0, 0));
	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->col_dim; j++)
		{
			tmp = mpf_get_prec(get_mpfmatrix_ij(mat, i, j));
			if(prec < tmp)
				prec = tmp;
		}
	}

	return prec;
}
#endif

void print_fmatrix(FMatrix mat)
{
	long int i, j;

	for(i = 0; i < mat->row_dim; i++)
	{
		printf("%5ld ", i);
		for(j = 0; j < mat->col_dim; j++)
			printf("%15.7e ", get_fmatrix_ij(mat, i, j));
		printf("\n");
	}
}

void print_dmatrix(DMatrix mat)
{
	long int i, j;

	for(i = 0; i < mat->row_dim; i++)
	{
		printf("%5ld ", i);
		for(j = 0; j < mat->col_dim; j++)
			printf("%25.17e ", get_dmatrix_ij(mat, i, j));
		printf("\n");
	}
}

#ifdef USE_GMP
void print_mpfmatrix(MPFMatrix mat)
{
	long int i, j;

	for(i = 0; i < mat->row_dim; i++)
	{
		printf("%5ld ", i);
		for(j = 0; j < mat->col_dim; j++)
		{
			mpf_out_str(stdout, 10, 0, get_mpfmatrix_ij(mat, i, j));
			printf(" ");
		}
		printf("\n");
	}
}
#endif

/*************************************************/
/* Vector Calculations for FVector               */
/*
void add_fvector(FVector c, FVector a, FVector b)
void sub_fvector(FVector c, FVector a, FVector b)
void cmul_fvector(FVector c, float val, FVector a)
float ip_fvector(FVector a, FVector b)
float norm1_fvector(FVector a)
float norm2_fvector(FVector a)
float normi_fvector(FVector a)
void subst_fvector(FVector c, FVector a)
*/
/*************************************************/
/* c = a + b */
void add_fvector(FVector c, FVector a, FVector b)
{
	long int i;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: add_fvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, get_fvector_i(a, i) + get_fvector_i(b, i));

}

/* c += a */
void add2_fvector(FVector c, FVector a)
{
	long int i;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: add2_fvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, get_fvector_i(c, i) + get_fvector_i(a, i));

}

/* c = a - b */
void sub_fvector(FVector c, FVector a, FVector b)
{
	long int i;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: sub_fvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, get_fvector_i(a, i) - get_fvector_i(b, i));

}

/* c -= a */
void sub2_fvector(FVector c, FVector a)
{
	long int i;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: sub2_fvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, get_fvector_i(c, i) - get_fvector_i(a, i));

}

/* c = val * a */
void cmul_fvector(FVector c, float val, FVector a)
{
	long int i;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: cmul_fvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, val * get_fvector_i(a, i));

}

/* c *= val */
void cmul2_fvector(FVector c, float val)
{
	long int i;

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, val * get_fvector_i(c, i));

}

/* (a, b) */
float ip_fvector(FVector a, FVector b)
{
	float tmp = 0.0;
	long int i;

	if(a->dim != b->dim)
	{
		fprintf(stderr, "ERROR: ip_fvector\n");
		return 0;
	}

	for(i = 0; i < a->dim; i++)
		tmp += get_fvector_i(a, i) * get_fvector_i(b, i);

	return tmp;
}

/* ||a||_1 */
float norm1_fvector(FVector a)
{
	float ret = 0.0;
	long int i;

	for(i = 0; i < a->dim; i++)
		ret += fabs((double)get_fvector_i(a, i));

	return ret;
}

/* ||a||_2 */
float norm2_fvector(FVector a)
{
	float ret = 0.0;
	long int i;

	for(i = 0; i < a->dim; i++)
		ret += get_fvector_i(a, i) * get_fvector_i(a, i);

	return sqrt(ret);
}

/* ||a||_infty */
float normi_fvector(FVector a)
{
	float ret, tmp;
	long int i;

	ret = fabs(get_fvector_i(a, 0));
	for(i = 1; i < a->dim; i++)
	{
		tmp = fabs((double)get_fvector_i(a, i));
		if(ret < tmp)
			ret = tmp;
	}

	return ret;
}

/* c := a */
void subst_fvector(FVector c, FVector a)
{
	long int i;

	for(i = 0; i < a->dim; i++)
		set_fvector_i(c, i, get_fvector_i(a, i));
}

/* c := 0 */
void set0_fvector(FVector c)
{
	long int i;

	for(i = 0; i < c->dim; i++)
		set_fvector_i(c, i, (float)0);
}

/*************************************************/
/* Vector Calculations for DVector               */
/*
void add_dvector(DVector c, DVector a, DVector b)
void add2_dvector(DVector c, DVector a)
void sub_dvector(DVector c, DVector a, DVector b)
void sub2_dvector(DVector c, DVector a)
void cmul_dvector(DVector c, double val, DVector a)
void cmul2_dvector(DVector c, double val)
void add_cmul_dvector(DVector c, DVector a, double val, DVector b)
double ip_dvector(DVector a, DVector b)
double norm1_dvector(DVector a)
double norm2_dvector(DVector a)
double normi_dvector(DVector a)
void subst_dvector(DVector c, DVector a)
*/
/*************************************************/
/* c = a + b */
void add_dvector(DVector c, DVector a, DVector b)
{
	long int i;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: add_dvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, get_dvector_i(a, i) + get_dvector_i(b, i));

}

/* c += a */
void add2_dvector(DVector c, DVector a)
{
	long int i;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: add2_dvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, get_dvector_i(c, i) + get_dvector_i(a, i));

}

/* c = a - b */
void sub_dvector(DVector c, DVector a, DVector b)
{
	long int i;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: sub_dvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, get_dvector_i(a, i) - get_dvector_i(b, i));

}

/* c -= a */
void sub2_dvector(DVector c, DVector a)
{
	long int i;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: sub2_dvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, get_dvector_i(c, i) - get_dvector_i(a, i));

}

/* c = val * a */
void cmul_dvector(DVector c, double val, DVector a)
{
	long int i;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: cmul_dvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, val * get_dvector_i(a, i));

}

/* c *= val */
void cmul2_dvector(DVector c, double val)
{
	long int i;

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, val * get_dvector_i(c, i));

}

/* c = a + val * b */
void add_cmul_dvector(DVector c, DVector a, double val, DVector b)
{
	long int i;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: add_cmul_dvector\n");
		return;
	}

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, get_dvector_i(a, i) + val * get_dvector_i(b, i));

}

/* (a, b) */
double ip_dvector(DVector a, DVector b)
{
	double tmp = 0.0;
	long int i;

	if(a->dim != b->dim)
	{
		fprintf(stderr, "ERROR: ip_dvector\n");
		return 0;
	}

	for(i = 0; i < a->dim; i++)
		tmp += get_dvector_i(a, i) * get_dvector_i(b, i);

	return tmp;
}

/* ||a||_1 */
double norm1_dvector(DVector a)
{
	double ret = 0.0;
	long int i;

	for(i = 0; i < a->dim; i++)
		ret += fabs(get_dvector_i(a, i));

	return ret;
}

/* ||a||_2 */
double norm2_dvector(DVector a)
{
	double ret = 0.0;
	long int i;

	for(i = 0; i < a->dim; i++)
		ret += get_dvector_i(a, i) * get_dvector_i(a, i);

	return sqrt(ret);
}

/* ||a||_infty */
double normi_dvector(DVector a)
{
	double ret, tmp;
	long int i;

	ret = fabs(get_dvector_i(a, 0));
	for(i = 1; i < a->dim; i++)
	{
		tmp = fabs(get_dvector_i(a, i));
		if(ret < tmp)
			ret = tmp;
	}

	return ret;
}

/* c := a */
void subst_dvector(DVector c, DVector a)
{
	long int i;

	for(i = 0; i < a->dim; i++)
		set_dvector_i(c, i, get_dvector_i(a, i));
}

/* c := 0 */
void set0_dvector(DVector c)
{
	long int i;

	for(i = 0; i < c->dim; i++)
		set_dvector_i(c, i, (double)0);
}

/* append 2005.07/12 */
/*
	ret(index_start) = src(src_index_start)
	 ...
	ret(index_end  ) = src(src_index_end)
*/
void copy_dvector_ij(DVector ret, long int index_start, long int index_end, DVector src, long int src_index_start, long int src_index_end)
{
	long int i, itmp;

	if((src_index_end - src_index_start) != (index_end - index_start))
	{
		fprintf(stderr, "Invalid index!(copy_dvector_ij)\n");
		return;
	}

	for(i = 0; i <= (index_end - index_start); i++)
	{
		set_dvector_i(ret, index_start + i, get_dvector_i(src, src_index_start + i));
//		printf("%d <----------------------------------> %d\n", index_start + i, src_index_start + i);
	}
}

/*************************************************/
/* Vector Calculations for MPFVector             */
/*
void add_mpfvector(MPFVector c, MPFVector a, MPFVector b)
void add2_mpfvector(MPFVector c, MPFVector a)
void sub_mpfvector(MPFVector c, MPFVector a, MPFVector b)
void sub2_mpfvector(MPFVector c, MPFVector a)
void cmul_mpfvector(MPFVector c, mpf_t val, MPFVector a)
void cmul2_mpfvector(MPFVector c, mpf_t val)
void add_cmul_mpfvector(MPFVector c, MPFVector a, mpf_t val, MPFVector b)
void ip_mpfvector(mpf_t ret, MPFVector a, MPFVector b)
void norm1_mpfvector(mpf_t ret, MPFVector a)
void norm2_mpfvector(mpf_t ret, MPFVector a)
void normi_mpfvector(mpf_t ret, MPFVector a)
void subst_mpfvector(MPFVector c, MPFVector a)
*/
/*************************************************/
#ifdef USE_GMP
/* c = a + b */
void add_mpfvector(MPFVector c, MPFVector a, MPFVector b)
{
	long int i;
	mpf_t tmp;
	unsigned long int prec;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: add_mpfvector\n");
		return;
	}

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_add(tmp, get_mpfvector_i(a, i), get_mpfvector_i(b, i));
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);

}

/* c += a */
void add2_mpfvector(MPFVector c, MPFVector a)
{
	long int i;
	mpf_t tmp;
	unsigned long int prec;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: add2_mpfvector\n");
		return;
	}

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_add(tmp, get_mpfvector_i(c, i), get_mpfvector_i(a, i));
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);
}

/* c = a - b */
void sub_mpfvector(MPFVector c, MPFVector a, MPFVector b)
{
	long int i;
	mpf_t tmp, tmpa, tmpb;
	unsigned long int prec;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: sub_mpfvector\n");
		return;
	}

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_sub(tmp, get_mpfvector_i(a, i), get_mpfvector_i(b, i));
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);
}

/* c -= a */
void sub2_mpfvector(MPFVector c, MPFVector a)
{
	long int i;
	mpf_t tmp;
	unsigned long int prec;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: sub2_mpfvector\n");
		return;
	}

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_sub(tmp, get_mpfvector_i(c, i), get_mpfvector_i(a, i));
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);

}

/* c = val * a */
void cmul_mpfvector(MPFVector c, mpf_t val, MPFVector a)
{
	long int i;
	mpf_t tmp, tmpa;
	unsigned long int prec;

	if(c->dim != a->dim)
	{
		fprintf(stderr, "ERROR: cmul_mpfvector\n");
		return;
	}

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_mul(tmp, val, get_mpfvector_i(a, i));
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);
}

/* c *= val */
void cmul2_mpfvector(MPFVector c, mpf_t val)
{
	long int i;
	mpf_t tmp;
	unsigned long int prec;

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_mul(tmp, val, get_mpfvector_i(c, i));
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);

}

/* c = a + val * b */
void add_cmul_mpfvector(MPFVector c, MPFVector a, mpf_t val, MPFVector b)
{
	long int i;
	mpf_t tmp, tmp2;
	unsigned long int prec;

	if((a->dim != b->dim) || (c->dim != a->dim) || (c->dim != b->dim))
	{
		fprintf(stderr, "ERROR: add_mpfvector\n");
		return;
	}

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);
	mpf_init2(tmp2, prec);

	for(i = 0; i < c->dim; i++)
	{
		mpf_mul(tmp2, val, get_mpfvector_i(b, i));
		mpf_add(tmp, get_mpfvector_i(a, i), tmp2);
		set_mpfvector_i(c, i, tmp);
	}

	mpf_clear(tmp);
	mpf_clear(tmp2);

}

/* ret = (a, b) */
void ip_mpfvector(mpf_t ret, MPFVector a, MPFVector b)
{
	long int i;
	mpf_t tmp;

	if(a->dim != b->dim)
	{
		fprintf(stderr, "ERROR: ip_mpfvector\n");
		return;
	}

	mpf_init2(tmp, mpf_get_prec(ret));

	mpf_set_ui(ret, 0UL); /* ret := 0 */
	for(i = 0; i < a->dim; i++)
	{
#ifndef USE_MPFR
		mpf_mul(tmp, get_mpfvector_i(a, i), get_mpfvector_i(b, i));
		mpf_add(ret, ret, tmp);
#else
		mpfr_fma(ret, get_mpfvector_i(a, i), get_mpfvector_i(b, i), ret, bnc_default_rounding_mode);
#endif
	}

	mpf_clear(tmp);
}

/* ret = ||a||_1 */
void norm1_mpfvector(mpf_t ret, MPFVector a)
{

	long int i;
	mpf_t tmp;

	mpf_init2(tmp, mpf_get_prec(ret));

	mpf_set_ui(ret, 0UL); /* ret := 0 */
	for(i = 0; i < a->dim; i++)
	{
		mpf_abs(tmp, get_mpfvector_i(a, i));
		mpf_add(ret, ret, tmp);
	}

	mpf_clear(tmp);
}

/* ret := ||a||_2 */
void norm2_mpfvector(mpf_t ret, MPFVector a)
{
	long int i;
	mpf_t tmp;

	mpf_init2(tmp, mpf_get_prec(ret));

	mpf_set_ui(ret, 0UL); /* ret := 0 */
	for(i = 0; i < a->dim; i++)
	{
#ifndef USE_MPFR
		mpf_mul(tmp, get_mpfvector_i(a, i), get_mpfvector_i(a, i));
		mpf_add(ret, ret, tmp);
#else
		mpfr_fma(ret, get_mpfvector_i(a, i), get_mpfvector_i(a, i), ret, bnc_default_rounding_mode);
#endif
	}
	mpf_sqrt(ret, ret);

	mpf_clear(tmp);
}

/* ||a||_infty */
void normi_mpfvector(mpf_t ret, MPFVector a)
{
	long int i;
	mpf_t tmp;

	mpf_init2(tmp, mpf_get_prec(ret));

	mpf_abs(ret, get_mpfvector_i(a, 0));
	for(i = 1; i < a->dim; i++)
	{
		mpf_abs(tmp, get_mpfvector_i(a, i));
		if(mpf_cmp(ret, tmp) < 0) /* ret < tmp */
			mpf_set(ret, tmp);
	}

	mpf_clear(tmp);
}

/* c := a */
void subst_mpfvector(MPFVector c, MPFVector a)
{
	long int i;

	for(i = 0; i < a->dim; i++)
		set_mpfvector_i(c, i, get_mpfvector_i(a, i));
}

/* c := 0 */
void set0_mpfvector(MPFVector c)
{
	unsigned long int prec;
	long int i;
	mpf_t tmp;

	prec = prec_mpfvector(c);

	mpf_init2(tmp, prec);
	mpf_set_ui(tmp, 0UL);

	for(i = 0; i < c->dim; i++)
		set_mpfvector_i(c, i, tmp);

	mpf_clear(tmp);
}

/* append 2005.07/12 */
/*
	ret(index_start) = src(src_index_start)
	 ...
	ret(index_end  ) = src(src_index_end)
*/
void copy_mpfvector_ij(MPFVector ret, long int index_start, long int index_end, MPFVector src, long int src_index_start, long int src_index_end)
{
	long int i, itmp;

	if((src_index_end - src_index_start) != (index_end - index_start))
	{
		fprintf(stderr, "Invalid index!(copy_mpfvector_ij)\n");
		return;
	}

	for(i = 0; i <= (index_end - index_start); i++)
	{
		set_mpfvector_i(ret, index_start + i, get_mpfvector_i(src, src_index_start + i));
//		printf("%d <----------------------------------> %d\n", index_start + i, src_index_start + i);
	}
}

#endif

/*************************************************/
/* Matrix Caluculations for FMatrix              */
/*
void add_fmatrix(FMatrix c, FMatrix a, FMatrix b);
void sub_fmatrix(FMatrix c, FMatrix a, FMatrix b);
void mul_fmatrix(FMatrix c, FMatrix a, FMatrix b);
void transpose_fmatrix(FMatrix c, FMatrix a);
void inv_fmatrix(FMatrix a);
void subst_fmatrux(FMatrix c, FMatrix a);
*/
/*************************************************/
/* c = a + b */
void add_fmatrix(FMatrix c, FMatrix a, FMatrix b)
{
	long int i, j, row_dim, col_dim;

	/* check row_dim */
	if((a->row_dim != b->row_dim) || (b->row_dim != c->row_dim) || (c->row_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: add_fmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if((a->col_dim != b->col_dim) || (b->col_dim != c->col_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: add_fmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
			set_fmatrix_ij(c, i, j, get_fmatrix_ij(a, i, j) + get_fmatrix_ij(b, i, j));
	}
}

/* c = a - b */
void sub_fmatrix(FMatrix c, FMatrix a, FMatrix b)
{
	long int i, j, row_dim, col_dim;

	/* check row_dim */
	if((a->row_dim != b->row_dim) || (b->row_dim != c->row_dim) || (c->row_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: add_fmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if((a->col_dim != b->col_dim) || (b->col_dim != c->col_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: add_fmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
			set_fmatrix_ij(c, i, j, get_fmatrix_ij(a, i, j) - get_fmatrix_ij(b, i, j));
	}
}


/* c = sc * a */
void cmul_fmatrix(FMatrix c, float sc, FMatrix a)
{
	long int i, j, row_dim, col_dim;

	/* check row_dim */
	if(a->row_dim != c->row_dim)
	{
		fprintf(stderr, "ERROR: cmul_fmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if(a->col_dim != c->col_dim)
	{
		fprintf(stderr, "ERROR: cmul_fmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
			set_fmatrix_ij(c, i, j, sc * get_fmatrix_ij(a, i, j));
	}
}

/* c = a * b */
void mul_fmatrix(FMatrix c, FMatrix a, FMatrix b)
{
	long int i, j, k;
	float tmp;

	/* dimension check */
	if((c->row_dim != a->row_dim) || (c->col_dim != b->row_dim) || (a->col_dim != b->row_dim))
	{
		fprintf(stderr, "ERROR: mul_fmatrix\n");
		return;
	}

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
		{
			tmp = 0.0;
			for(k = 0; k < a->col_dim; k++)
				tmp += get_fmatrix_ij(a, i, k) * get_fmatrix_ij(b, k, j);
			set_fmatrix_ij(c, i, j, tmp);
		}
	}
}

/* c = a^T */
void transpose_fmatrix(FMatrix c, FMatrix a)
{
	long int i, j;

	/* Check Dimentions */
	if((c->row_dim != a->col_dim) || (c->col_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: transpose_fmatrix\n");
		return;
	}
	
	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_fmatrix_ij(c, i, j, get_fmatrix_ij(a, j, i));
	}
}

/* c := a */
void subst_fmatrix(FMatrix c, FMatrix a)
{
	long int i, j;
	float tmp;

	if((c->row_dim != a->row_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: subst_fmatrix\n");
		return;
	}

	for(i = 0; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			set_fmatrix_ij(c, i, j, get_fmatrix_ij(a, i, j));
		}
	}
}

/* c := 0 */
void set0_fmatrix(FMatrix c)
{
	long int i, j;

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_fmatrix_ij(c, i, j, 0.0);
	}
}

/* c := I */
void setI_fmatrix(FMatrix c)
{
	long int i, j;

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_fmatrix_ij(c, i, j, 0.0);
		if(i < c->col_dim)
			set_fmatrix_ij(c, i, i, 1.0);
	}
}

/* v = a * vb */
void mul_fmatrix_fvec(FVector v, FMatrix a, FVector vb)
{
	long int i, j;
	float tmp;

	/* Check Dimension */
	if((v->dim != a->row_dim) || (vb->dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: mul_fmatrix_fvec\n");
		return;
	}
	
	for(i = 0; i < a->row_dim; i++)
	{
		tmp = 0.0;
		for(j = 0; j < a->col_dim; j++)
			tmp += get_fmatrix_ij(a, i, j) * get_fvector_i(vb, j);
		set_fvector_i(v, i, tmp);
	}
}

/* a = a^(-1) */
/* square matrix only */
void inv_fmatrix(FMatrix a)
{
	long int i, j, k, dim;
	float tmp, aii;

	/* Check Dimensions */
	if(a->row_dim != a->col_dim)
	{
		fprintf(stderr, "ERROR: inv_fmatrix\n");
		return;
	}

	dim = a->row_dim;

	for(i = 0; i < dim; i++)
	{
		if(get_fmatrix_ij(a, i, i) == 0.0)
		{
			fprintf(stderr, "ERROR: inv_fmatrix: Pivot(%ld,%ld) is zero.\n", i, i);
			return;
		}

		aii = 1.0 / get_fmatrix_ij(a, i, i);
		set_fmatrix_ij(a, i, i, aii);

		for(j = 0; j < i; j++)
			set_fmatrix_ij(a, i, j, get_fmatrix_ij(a, i, j)* aii);
		for(j = i + 1; j < dim; j++)
			set_fmatrix_ij(a, i, j, get_fmatrix_ij(a, i, j) * aii);

		for(j = 0; j < i; j++)
		{
			for(k = 0; k < i; k++)
				set_fmatrix_ij(a, j, k, get_fmatrix_ij(a, j, k) - get_fmatrix_ij(a, j, i) * get_fmatrix_ij(a, i, k));
			for(k = i + 1; k < dim; k++)
				set_fmatrix_ij(a, j, k, get_fmatrix_ij(a, j, k) - get_fmatrix_ij(a, j, i) * get_fmatrix_ij(a, i, k));
		}

		for(j = i + 1; j < dim; j++)
		{
			for(k = 0; k < i; k++)
				set_fmatrix_ij(a, j, k, get_fmatrix_ij(a, j, k) - get_fmatrix_ij(a, j, i) * get_fmatrix_ij(a, i, k));
			for(k = i + 1; k < dim; k++)
				set_fmatrix_ij(a, j, k, get_fmatrix_ij(a, j, k) - get_fmatrix_ij(a, j, i) * get_fmatrix_ij(a, i, k));
		}

		for(j = 0; j < i; j++)
			set_fmatrix_ij(a, j, i, get_fmatrix_ij(a, j, i) * -aii);
		for(j = i + 1; j < dim; j++)
			set_fmatrix_ij(a, j, i, get_fmatrix_ij(a, j, i) * -aii);	}
}

/*************************************************/
/* Matrix Caluculations for DMatrix              */
/*
double normf_dmatrix(DMatrix mat)
double normi_dmatrix(DMatrix mat)
double norm1_dmatrix(DMatrix mat)
void add_dmatrix(DMatrix c, DMatrix a, DMatrix b);
void sub_dmatrix(DMatrix c, DMatrix a, DMatrix b);
void mul_dmatrix(DMatrix c, DMatrix a, DMatrix b);
void transpose_dmatrix(DMatrix c, DMatrix a);
void mul_dmatrix_dvec(DVector v, DMatrix a, DVector vb)
void mul_dmatrixt_dvec(DVector v, DMatrix a, DVector vb)
void inv_dmatrix(DMatrix a);
void subst_dmatrux(DMatrix c, DMatrix a);
*/
/*************************************************/
/* Frobenius Norm of Matrix */
double normf_dmatrix(DMatrix mat)
{
	long int i, j;
	double ret;

	ret = 0.0;
	for(i = 0; i < mat->row_dim; i++)
		for(j = 0; j < mat->col_dim; j++)
			ret += (get_dmatrix_ij(mat, i, j) * get_dmatrix_ij(mat, i, j));

	ret = sqrt(ret);

	return ret;
}

/* Infinity Norm of Matrix */
double normi_dmatrix(DMatrix mat)
{
	long int i, j;
	double ret, sum;

	ret = 0.0;
	for(i = 0; i < mat->row_dim; i++)
	{
		sum = 0.0;
		for(j = 0; j < mat->col_dim; j++)
			sum += fabs(get_dmatrix_ij(mat, i, j));
		if(ret < sum)
			ret = sum;
	}

	return ret;
}

/* 1 Norm of Matrix */
double norm1_dmatrix(DMatrix mat)
{
	long int i, j;
	double ret, sum;

	ret = 0.0;
	for(j = 0; j < mat->col_dim; j++)
	{
		sum = 0.0;
		for(i = 0; i < mat->row_dim; i++)
			sum += fabs(get_dmatrix_ij(mat, i, j));
		if(ret < sum)
			ret = sum;
	}

	return ret;
}

/* c = a + b */
void add_dmatrix(DMatrix c, DMatrix a, DMatrix b)
{
	long int i, j, row_dim, col_dim;

	/* check row_dim */
	if((a->row_dim != b->row_dim) || (b->row_dim != c->row_dim) || (c->row_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: add_dmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if((a->col_dim != b->col_dim) || (b->col_dim != c->col_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: add_dmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
			set_dmatrix_ij(c, i, j, get_dmatrix_ij(a, i, j) + get_dmatrix_ij(b, i, j));
	}
}

/* c = a - b */
void sub_dmatrix(DMatrix c, DMatrix a, DMatrix b)
{
	long int i, j, row_dim, col_dim;

	/* check row_dim */
	if((a->row_dim != b->row_dim) || (b->row_dim != c->row_dim) || (c->row_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: sub_dmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if((a->col_dim != b->col_dim) || (b->col_dim != c->col_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: sub_dmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
			set_dmatrix_ij(c, i, j, get_dmatrix_ij(a, i, j) - get_dmatrix_ij(b, i, j));
	}
}

/* c = sc * a */
void cmul_dmatrix(DMatrix c, double sc, DMatrix a)
{
	long int i, j, row_dim, col_dim;

	/* check row_dim */
	if(a->row_dim != c->row_dim)
	{
		fprintf(stderr, "ERROR: cmul_dmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if(a->col_dim != c->col_dim)
	{
		fprintf(stderr, "ERROR: cmul_dmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
			set_dmatrix_ij(c, i, j, sc * get_dmatrix_ij(a, i, j));
	}
}

/* c = a * b */
void mul_dmatrix(DMatrix c, DMatrix a, DMatrix b)
{
	long int i, j, k;
	double tmp;

	/* dimension check */
	if((c->row_dim != a->row_dim) || (c->col_dim != b->row_dim) || (a->col_dim != b->row_dim))
	{
		fprintf(stderr, "ERROR: mul_dmatrix\n");
		return;
	}

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
		{
			tmp = 0.0;
			for(k = 0; k < a->col_dim; k++)
				tmp += get_dmatrix_ij(a, i, k) * get_dmatrix_ij(b, k, j);
			set_dmatrix_ij(c, i, j, tmp);
		}
	}
}

/* c = a^T */
void transpose_dmatrix(DMatrix c, DMatrix a)
{
	long int i, j;

	/* Check Dimentions */
	if((c->row_dim != a->col_dim) || (c->col_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: transpose_dmatrix\n");
		return;
	}
	
	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_dmatrix_ij(c, i, j, get_dmatrix_ij(a, j, i));
	}
}

/* c := a */
void subst_dmatrix(DMatrix c, DMatrix a)
{
	long int i, j;

	if((c->row_dim != a->row_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: subst_dmatrix\n");
		return;
	}

	for(i = 0; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			set_dmatrix_ij(c, i, j, get_dmatrix_ij(a, i, j));
		}
	}
}

/* c := 0 */
void set0_dmatrix(DMatrix c)
{
	long int i, j;

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_dmatrix_ij(c, i, j, 0.0);
	}
}

/* c := I */
void setI_dmatrix(DMatrix c)
{
	long int i, j;

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_dmatrix_ij(c, i, j, 0.0);
		if(i < c->col_dim)
			set_dmatrix_ij(c, i, i, 1.0);
	}
}

/* v = a * vb */
void mul_dmatrix_dvec(DVector v, DMatrix a, DVector vb)
{
	long int i, j;
	double tmp;

	/* Check Dimension */
	if((v->dim != a->row_dim) || (vb->dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: mul_dmatrix_dvec\n");
		return;
	}
	
	for(i = 0; i < a->row_dim; i++)
	{
		tmp = 0.0;
		for(j = 0; j < a->col_dim; j++)
			tmp += get_dmatrix_ij(a, i, j) * get_dvector_i(vb, j);
		set_dvector_i(v, i, tmp);
	}
}

/* v = a^T * vb */
void mul_dmatrixt_dvec(DVector v, DMatrix a, DVector vb)
{
	long int i, j;
	double tmp;

	/* Check Dimension */
	if((v->dim != a->col_dim) || (vb->dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: mul_dmatrixt_dvec\n");
		return;
	}
	
	for(i = 0; i < a->col_dim; i++)
	{
		tmp = 0.0;
		for(j = 0; j < a->row_dim; j++)
			tmp += get_dmatrix_ij(a, j, i) * get_dvector_i(vb, j);
		set_dvector_i(v, i, tmp);
	}
}

/* a = a^(-1) */
/* square matrix only */
void inv_dmatrix(DMatrix a)
{
	long int i, j, k, dim;
	double tmp, aii;

	/* Check Dimensions */
	if(a->row_dim != a->col_dim)
	{
		fprintf(stderr, "ERROR: inv_dmatrix\n");
		return;
	}

	dim = a->row_dim;

	for(i = 0; i < dim; i++)
	{
		if(get_dmatrix_ij(a, i, i) == 0.0)
		{
			fprintf(stderr, "ERROR: inv_dmatrix: Pivot(%ld,%ld) is zero.\n", i, i);
			return;
		}

		aii = 1.0 / get_dmatrix_ij(a, i, i);
		set_dmatrix_ij(a, i, i, aii);

		for(j = 0; j < i; j++)
			set_dmatrix_ij(a, i, j, get_dmatrix_ij(a, i, j)* aii);
		for(j = i + 1; j < dim; j++)
			set_dmatrix_ij(a, i, j, get_dmatrix_ij(a, i, j) * aii);

		for(j = 0; j < i; j++)
		{
			for(k = 0; k < i; k++)
				set_dmatrix_ij(a, j, k, get_dmatrix_ij(a, j, k) - get_dmatrix_ij(a, j, i) * get_dmatrix_ij(a, i, k));
			for(k = i + 1; k < dim; k++)
				set_dmatrix_ij(a, j, k, get_dmatrix_ij(a, j, k) - get_dmatrix_ij(a, j, i) * get_dmatrix_ij(a, i, k));
		}

		for(j = i + 1; j < dim; j++)
		{
			for(k = 0; k < i; k++)
				set_dmatrix_ij(a, j, k, get_dmatrix_ij(a, j, k) - get_dmatrix_ij(a, j, i) * get_dmatrix_ij(a, i, k));
			for(k = i + 1; k < dim; k++)
				set_dmatrix_ij(a, j, k, get_dmatrix_ij(a, j, k) - get_dmatrix_ij(a, j, i) * get_dmatrix_ij(a, i, k));
		}

		for(j = 0; j < i; j++)
			set_dmatrix_ij(a, j, i, get_dmatrix_ij(a, j, i) * -aii);
		for(j = i + 1; j < dim; j++)
			set_dmatrix_ij(a, j, i, get_dmatrix_ij(a, j, i) * -aii);	}
}

/*************************************************/
/* Matrix Caluculations for MPFMatrix            */
/*
void normf_mpfmatrix(mpf_t ret, MPFMatrix mat)
void norm1_mpfmatrix(mpf_t ret, MPFMatrix mat)
void normi_mpfmatrix(mpf_t ret, MPFMatrix mat)
void add_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b);
void sub_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b);
void mul_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b);
void mul_mpfmatrix_mpfvec(MPFVector v, MPFMatrix a, MPFVector vb)
void mul_mpfmatrixt_mpfvec(MPFVector v, MPFMatrix a, MPFVector vb)
void transpose_mpfmatrix(MPFMatrix c, MPFMatrix a);
void inv_mpfmatrix(MPFMatrix a);
void subst_mpfmatrux(MPFMatrix c, MPFMatrix a);
*/
/*************************************************/
#ifdef USE_GMP
/* Frobenius Norm of Matrix */
void normf_mpfmatrix(mpf_t ret, MPFMatrix mat)
{
	long int i, j;
	mpf_t tmp;

	mpf_init2(tmp, mat->prec);
	mpf_set_ui(ret, 0UL);
	for(i = 0; i < mat->row_dim; i++)
		for(j = 0; j < mat->col_dim; j++)
		{
			mpf_mul(tmp, get_mpfmatrix_ij(mat, i, j), get_mpfmatrix_ij(mat, i, j));
			mpf_add(ret, ret, tmp);
		}

	mpf_sqrt(ret, ret);

	mpf_clear(tmp);

	return;
}

/* Infinity Norm of Matrix */
void normi_mpfmatrix(mpf_t ret, MPFMatrix mat)
{
	long int i, j;
	mpf_t tmp, sum;

	mpf_init2(tmp, mat->prec);
	mpf_init2(sum, mat->prec);

	mpf_set_ui(ret, 0UL);
	for(i = 0; i < mat->row_dim; i++)
	{
		mpf_set_ui(sum, 0UL);
		for(j = 0; j < mat->col_dim; j++)
		{
			mpf_abs(tmp, get_mpfmatrix_ij(mat, i, j));
			mpf_add(sum, sum, tmp);
		}
		if(mpf_cmp(ret, sum) < 0)
			mpf_set(ret, sum);
	}

	mpf_clear(tmp);
	mpf_clear(sum);

	return;
}

/* 1 Norm of Matrix */
void norm1_mpfmatrix(mpf_t ret, MPFMatrix mat)
{
	long int i, j;
	mpf_t tmp, sum;

	mpf_init2(tmp, mat->prec);
	mpf_init2(sum, mat->prec);

	mpf_set_ui(ret, 0UL);
	for(j = 0; j < mat->col_dim; j++)
	{
		mpf_set_ui(sum, 0UL);
		for(i = 0; i < mat->row_dim; i++)
		{
			mpf_abs(tmp, get_mpfmatrix_ij(mat, i, j));
			mpf_add(sum, sum, tmp);
		}
		if(mpf_cmp(ret, sum) < 0)
			mpf_set(ret, sum);
	}

	mpf_clear(tmp);
	mpf_clear(sum);

	return;
}

/* c := a + b */
void add_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b)
{
	long int i, j, row_dim, col_dim;
	mpf_t tmp;

	/* check row_dim */
	if((a->row_dim != b->row_dim) || (b->row_dim != c->row_dim) || (c->row_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: add_mpfmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if((a->col_dim != b->col_dim) || (b->col_dim != c->col_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: add_mpfmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	mpf_init2(tmp, prec_mpfmatrix(c));

	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
		{
			mpf_add(tmp, get_mpfmatrix_ij(a, i, j), get_mpfmatrix_ij(b, i, j));
			set_mpfmatrix_ij(c, i, j, tmp);
		}
	}

	mpf_clear(tmp);
}

/* c := a - b */
void sub_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b)
{
	long int i, j, row_dim, col_dim;
	mpf_t tmp;

	/* check row_dim */
	if((a->row_dim != b->row_dim) || (b->row_dim != c->row_dim) || (c->row_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: add_mpfmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if((a->col_dim != b->col_dim) || (b->col_dim != c->col_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: add_mpfmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	mpf_init2(tmp, prec_mpfmatrix(c));
	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
		{
			mpf_sub(tmp, get_mpfmatrix_ij(a, i, j), get_mpfmatrix_ij(b, i, j));
			set_mpfmatrix_ij(c, i, j, tmp);
		}
	}
	mpf_clear(tmp);
}

/* c := sc * a */
void cmul_mpfmatrix(MPFMatrix c, mpf_t sc, MPFMatrix a)
{
	long int i, j, row_dim, col_dim;
	mpf_t tmp;

	/* check row_dim */
	if(a->row_dim != c->row_dim)
	{
		fprintf(stderr, "ERROR: cmul_mpfmatrix\n");
		return;
	}
	row_dim = c->row_dim;

	/* check col_dim */
	if(a->col_dim != c->col_dim)
	{
		fprintf(stderr, "ERROR: cmul_mpfmatrix\n");
		return;
	}
	col_dim = c->col_dim;

	mpf_init2(tmp, prec_mpfmatrix(c));
	for(i = 0; i < row_dim; i++)
	{
		for(j = 0; j < col_dim; j++)
		{
			mpf_mul(tmp, sc, get_mpfmatrix_ij(a, i, j));
			set_mpfmatrix_ij(c, i, j, tmp);
		}
	}
	mpf_clear(tmp);
}


/* c = a * b */
void mul_mpfmatrix(MPFMatrix c, MPFMatrix a, MPFMatrix b)
{
	long int i, j, k;
	mpf_t tmp, tmp1;

	/* dimension check */
	if((c->row_dim != a->row_dim) || (c->col_dim != b->row_dim) || (a->col_dim != b->row_dim))
	{
		fprintf(stderr, "ERROR: mul_mpfmatrix\n");
		return;
	}

	mpf_init2(tmp, prec_mpfmatrix(c));
	mpf_init2(tmp1, prec_mpfmatrix(c));
	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
		{
			mpf_set_ui(tmp, 0UL);
			for(k = 0; k < a->col_dim; k++)
			{
#ifndef USE_MPFR
				mpf_mul(tmp1, get_mpfmatrix_ij(a, i, k), get_mpfmatrix_ij(b, k, j));
				mpf_add(tmp, tmp, tmp1);
#else
				mpfr_fma(tmp, get_mpfmatrix_ij(a, i, k), get_mpfmatrix_ij(b, k, j), tmp, bnc_default_rounding_mode);
#endif
			}
			set_mpfmatrix_ij(c, i, j, tmp);
		}
	}
	mpf_clear(tmp);
	mpf_clear(tmp1);
}

/* c = a^T */
void transpose_mpfmatrix(MPFMatrix c, MPFMatrix a)
{
	long int i, j;

	/* Check Dimensions */
	if((c->row_dim != a->col_dim) || (c->col_dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: transpose_mpfmatrix\n");
		return;
	}
	
	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_mpfmatrix_ij(c, i, j, get_mpfmatrix_ij(a, j, i));
	}
}

/* c := a */
void subst_mpfmatrix(MPFMatrix c, MPFMatrix a)
{
	long int i, j;

	if((c->row_dim != a->row_dim) || (c->col_dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: subst_mpfmatrix\n");
		return;
	}

	for(i = 0; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
			set_mpfmatrix_ij(c, i, j, get_mpfmatrix_ij(a, i, j));
	}
}

/* c := 0 */
void set0_mpfmatrix(MPFMatrix c)
{
	long int i, j;
	mpf_t tmp;

	mpf_init2(tmp, c->prec);
	mpf_set_ui(tmp, 0UL);

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_mpfmatrix_ij(c, i, j, tmp);
	}

	mpf_clear(tmp);
}

/* c := I */
void setI_mpfmatrix(MPFMatrix c)
{
	long int i, j;
	mpf_t tmp0, tmp1;

	mpf_init2(tmp0, c->prec);
	mpf_init2(tmp1, c->prec);
	mpf_set_ui(tmp0, 0UL);
	mpf_set_ui(tmp1, 1UL);

	for(i = 0; i < c->row_dim; i++)
	{
		for(j = 0; j < c->col_dim; j++)
			set_mpfmatrix_ij(c, i, j, tmp0);
		if(i < c->col_dim)
			set_mpfmatrix_ij(c, i, i, tmp1);
	}

	mpf_clear(tmp0);
	mpf_clear(tmp1);
}

/* v := a * vb */
void mul_mpfmatrix_mpfvec(MPFVector v, MPFMatrix a, MPFVector vb)
{
	long int i, j;
	mpf_t tmp, tmp1;

	/* Check Dimension */
	if((v->dim != a->row_dim) || (vb->dim != a->col_dim))
	{
		fprintf(stderr, "ERROR: mul_mpfmatrix_dvec\n");
		return;
	}

	mpf_init2(tmp, prec_mpfvector(v));
	mpf_init2(tmp1, prec_mpfvector(v));
	for(i = 0; i < a->row_dim; i++)
	{
		mpf_set_ui(tmp, 0UL);
		for(j = 0; j < a->col_dim; j++)
		{
#ifndef USE_MPFR
			mpf_mul(tmp1, get_mpfmatrix_ij(a, i, j), get_mpfvector_i(vb, j));
			mpf_add(tmp, tmp, tmp1);
#else
			mpfr_fma(tmp, get_mpfmatrix_ij(a, i, j), get_mpfvector_i(vb, j), tmp, bnc_default_rounding_mode);
#endif
		}
		set_mpfvector_i(v, i, tmp);
	}
	mpf_clear(tmp);
	mpf_clear(tmp1);
}

/* v := a^T * vb */
void mul_mpfmatrixt_mpfvec(MPFVector v, MPFMatrix a, MPFVector vb)
{
	long int i, j;
	mpf_t tmp, tmp1;

	/* Check Dimension */
	if((v->dim != a->col_dim) || (vb->dim != a->row_dim))
	{
		fprintf(stderr, "ERROR: mul_mpfmatrixt_mpfvec\n");
		return;
	}

	mpf_init2(tmp, prec_mpfvector(v));
	mpf_init2(tmp1, prec_mpfvector(v));
	for(i = 0; i < a->col_dim; i++)
	{
		mpf_set_ui(tmp, 0UL);
		for(j = 0; j < a->row_dim; j++)
		{
#ifndef USE_MPFR
			mpf_mul(tmp1, get_mpfmatrix_ij(a, j, i), get_mpfvector_i(vb, j));
			mpf_add(tmp, tmp, tmp1);
#else
			mpfr_fma(tmp, get_mpfmatrix_ij(a, j, i), get_mpfvector_i(vb, j), tmp, bnc_default_rounding_mode);
#endif
		}
		set_mpfvector_i(v, i, tmp);
	}
	mpf_clear(tmp);
	mpf_clear(tmp1);
}

/* a = a^(-1) */
/* square matrix only */
void inv_mpfmatrix(MPFMatrix a)
{
	long int i, j, k, dim;
	mpf_t tmp, aii;

	/* Check Dimensions */
	if(a->row_dim != a->col_dim)
	{
		fprintf(stderr, "ERROR: inv_mpfmatrix\n");
		return;
	}

	mpf_init2(tmp, prec_mpfmatrix(a));
	mpf_init2(aii, prec_mpfmatrix(a));
	dim = a->row_dim;

	for(i = 0; i < dim; i++)
	{
		if(mpf_cmp_ui(get_mpfmatrix_ij(a, i, i), 0UL) == 0) 
		{
			fprintf(stderr, "ERROR: inv_mpfmatrix: Pivot(%ld,%ld) is zero.\n", i, i);
			return;
		}

		mpf_ui_div(aii, 1UL, get_mpfmatrix_ij(a, i, i));
		set_mpfmatrix_ij(a, i, i, aii);

		for(j = 0; j < i; j++)
		{
			mpf_mul(tmp, get_mpfmatrix_ij(a, i, j), aii);
			set_mpfmatrix_ij(a, i, j, tmp);
		}
		for(j = i + 1; j < dim; j++)
		{
			mpf_mul(tmp, get_mpfmatrix_ij(a, i, j), aii);
			set_mpfmatrix_ij(a, i, j, tmp);
		}

		for(j = 0; j < i; j++)
		{
			for(k = 0; k < i; k++)
			{
				mpf_mul(tmp, get_mpfmatrix_ij(a, j, i), get_mpfmatrix_ij(a, i, k));
				mpf_sub(tmp, get_mpfmatrix_ij(a, j, k), tmp);
				set_mpfmatrix_ij(a, j, k, tmp);
			}
			for(k = i + 1; k < dim; k++)
			{
				mpf_mul(tmp, get_mpfmatrix_ij(a, j, i), get_mpfmatrix_ij(a, i, k));
				mpf_sub(tmp, get_mpfmatrix_ij(a, j, k), tmp);
				set_mpfmatrix_ij(a, j, k, tmp);
			}
		}

		for(j = i + 1; j < dim; j++)
		{
			for(k = 0; k < i; k++)
			{
				mpf_mul(tmp, get_mpfmatrix_ij(a, j, i), get_mpfmatrix_ij(a, i, k));
				mpf_sub(tmp, get_mpfmatrix_ij(a, j, k), tmp);
				set_mpfmatrix_ij(a, j, k, tmp);
			}
			for(k = i + 1; k < dim; k++)
			{
				mpf_mul(tmp, get_mpfmatrix_ij(a, j, i), get_mpfmatrix_ij(a, i, k));
				mpf_sub(tmp, get_mpfmatrix_ij(a, j, k), tmp);
				set_mpfmatrix_ij(a, j, k, tmp);
			}
		}

		for(j = 0; j < i; j++)
		{
			mpf_neg(tmp, aii); /* tmp := -aii */
			mpf_mul(tmp, tmp, get_mpfmatrix_ij(a, j, i));
			set_mpfmatrix_ij(a, j, i, tmp);
		}
		for(j = i + 1; j < dim; j++)
		{
			mpf_neg(tmp, aii); /* tmp := -aii */
			mpf_mul(tmp, tmp, get_mpfmatrix_ij(a, j, i));
			set_mpfmatrix_ij(a, j, i, tmp);
		}
	}

	mpf_clear(aii);
	mpf_clear(tmp);
}
#endif

