/**********************************************/
/* test_cg.c:                                 */
/* Copyright (C) 2003 Tomonori Kouya          */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
/*************************************************/
/* Test Program for FCG, DCG, MPFCG              */
/*************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"


void get_fproblem(FMatrix a, FVector b, FVector ans)
{
	long int i, j, k;
	float tmp;

	/* Frank Matrix */
	for(i = 0; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			if(i < j)
				set_fmatrix_ij(a, i, j, (float)(a->col_dim - j));
			else
				set_fmatrix_ij(a, i, j, (float)(a->col_dim - i));
		}
	}

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_fvector_i(ans, i, (float)i);

	/* Make constant vector */
	mul_fmatrix_fvec(b, a, ans);
}

void get_dproblem(DMatrix a, DVector b, DVector ans)
{
	long int i, j, k;
	double tmp;

	/* Frank Matrix */
	for(i = 0; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			if(i < j)
				set_dmatrix_ij(a, i, j, (double)(a->col_dim - j));
			else
				set_dmatrix_ij(a, i, j, (double)(a->col_dim - i));
		}
	}

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_dvector_i(ans, i, (double)i);

	/* Make constant vector */
	mul_dmatrix_dvec(b, a, ans);
}

#ifdef USE_GMP
void get_mpfproblem(MPFMatrix a, MPFVector b, MPFVector ans)
{
	long int i, j, k;
	mpf_t tmp;

	mpf_init2(tmp, prec_mpfvector(ans));

	/* Frank Matrix */
	for(i = 0; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			if(i < j)
			{
				mpf_set_si(tmp, a->col_dim - j);
				set_mpfmatrix_ij(a, i, j, tmp);
			}
			else
			{
				mpf_set_si(tmp, a->col_dim - i);
				set_mpfmatrix_ij(a, i, j, tmp);
			}
		}
	}

	/* Answer */
	for(i = 0; i < ans->dim; i++)
	{
		mpf_set_si(tmp, i);
		set_mpfvector_i(ans, i, tmp);
	}

	/* Make constant vector */
	mul_mpfmatrix_mpfvec(b, a, ans);
}
#endif

#define DIM 100

main()
{
	FMatrix fa;
	FVector fb, fx, fans;
	DMatrix da;
	DVector db, dx, dans;
	double start, ftime, dtime;
#ifdef USE_GMP
	MPFMatrix mpfa;
	MPFVector mpfb, mpfx, mpfans;
	mpf_t reps, aeps;
	MPFMatrix mpfa2;
	MPFVector mpfb2, mpfx2, mpfans2;
	mpf_t reps2, aeps2;
	MPFMatrix mpfa3;
	MPFVector mpfb3, mpfx3, mpfans3;
	mpf_t reps3, aeps3;
	long int itimes_mpf, itimes_mpf2, itimes_mpf3;
	double mpftime[3];
#endif
	long int itimes_f, itimes_d;
	long int i, j;

/* Float */
	/* initialize */
	fa = init_fmatrix(DIM, DIM);
	fb = init_fvector(DIM);
	fx = init_fvector(DIM);
	fans = init_fvector(DIM);

	/* get problem */
	get_fproblem(fa, fb, fans);

	print_fmatrix(fa);

	/* run FCG */
	start = get_secv();
	itimes_f = FCG(fx, fa, fb, 1.0e-5, 1.0e-50, DIM * 5);
	ftime = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %15.7e %15.7e\n", i, get_fvector_i(fx, i), get_fvector_i(fans, i));
		
	/* end */
	free_fmatrix(fa);
	free_fvector(fb);
	free_fvector(fx);
	free_fvector(fans);

/* Double */
	/* initialize */
	da = init_dmatrix(DIM, DIM);
	db = init_dvector(DIM);
	dx = init_dvector(DIM);
	dans = init_dvector(DIM);

	/* get problem */
	get_dproblem(da, db, dans);

	print_dmatrix(da);

	/* run DCG */
	start = get_secv();
	itimes_d = DCG(dx, da, db, 1.0e-13, 1.0e-99, DIM * 5);
	dtime = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));
		
	/* end */
	free_dmatrix(da);
	free_dvector(db);
	free_dvector(dx);
	free_dvector(dans);

#ifdef USE_GMP
/* MPF */
	set_bnc_default_prec(128);
	/* initialize */
	mpf_init(reps);
	mpf_init2(reps2, 256);
	mpf_init2(reps3, 512);
	mpf_init(aeps);
	mpf_init2(aeps2, 256);
	mpf_init2(aeps3, 512);
	mpfa  = init_mpfmatrix(DIM, DIM);
	mpfa2 = init2_mpfmatrix(DIM, DIM, 256);
	mpfa3 = init2_mpfmatrix(DIM, DIM, 512);
	mpfb  = init_mpfvector(DIM);
	mpfb2 = init2_mpfvector(DIM, 256);
	mpfb3 = init2_mpfvector(DIM, 512);
	mpfx  = init_mpfvector(DIM);
	mpfx2 = init2_mpfvector(DIM, 256);
	mpfx3 = init2_mpfvector(DIM, 512);
	mpfans  = init_mpfvector(DIM);
	mpfans2 = init2_mpfvector(DIM, 256);
	mpfans3 = init2_mpfvector(DIM, 512);

	/* get problem */
	get_mpfproblem(mpfa, mpfb, mpfans);
	get_mpfproblem(mpfa2, mpfb2, mpfans2);
	get_mpfproblem(mpfa3, mpfb3, mpfans3);

	print_mpfmatrix(mpfa);
	print_mpfmatrix(mpfa2);
	print_mpfmatrix(mpfa3);

	/* run MPFFCG */
	mpf_set_d(reps, 1.0e-20);
	mpf_set_d(reps2, 1.0e-20);
	mpf_set_d(reps3, 1.0e-20);
	mpf_set_d(aeps, 1.0e-50);
	mpf_set_d(aeps2, 1.0e-50);
	mpf_set_d(aeps3, 1.0e-50);
	start = get_secv();
	itimes_mpf  = MPFCG(mpfx, mpfa, mpfb, reps, aeps, DIM * 5);
	mpftime[0] = get_secv() - start;

	start = get_secv();
	itimes_mpf2 = MPFCG(mpfx2, mpfa2, mpfb2, reps2, aeps2, DIM * 5);
	mpftime[1] = get_secv() - start;

	start = get_secv();
	itimes_mpf3 = MPFCG(mpfx3, mpfa3, mpfb3, reps3, aeps3, DIM * 5);
	mpftime[2] = get_secv() - start;

	/* print */
	for(i = 0; i < DIM; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx2, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx3, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}

	/* end */
	mpf_clear(reps); mpf_clear(aeps);
	mpf_clear(reps2); mpf_clear(aeps2);
	mpf_clear(reps3); mpf_clear(aeps3);
	free_mpfmatrix(mpfa);
	free_mpfmatrix(mpfa2);
	free_mpfmatrix(mpfa3);
	free_mpfvector(mpfb);
	free_mpfvector(mpfb2);
	free_mpfvector(mpfb3);
	free_mpfvector(mpfx);
	free_mpfvector(mpfx2);
	free_mpfvector(mpfx3);
	free_mpfvector(mpfans);
	free_mpfvector(mpfans2);
	free_mpfvector(mpfans3);
#endif

	/* print itimes */
	printf("Iterative Times\n");
	printf("single    : %ld(%f)\n", itimes_f, ftime);
	printf("double    : %ld(%f)\n", itimes_d, dtime);
#ifdef USE_GMP
	printf("mpf_t(128): %ld(%f)\n", itimes_mpf, mpftime[0]);
	printf("mpf_t(256): %ld(%f)\n", itimes_mpf2, mpftime[1]);
	printf("mpf_t(512): %ld(%f)\n", itimes_mpf3, mpftime[2]);
#endif
}

